from typing import Any
import torch
import torchvision.transforms.functional as tf
import numpy as np


class Resize:
    """Resize the image and coordinate labels."""

    def __init__(self, new_size=(350, 350)):
        self.new_width = new_size[0]
        self.new_height = new_size[1]

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        f_x, f_y = f_label
        od_x, od_y = od_label
        original_width, original_height = image.size
        image_new = tf.resize(image, (self.new_height, self.new_width))

        f_x_new = f_x * self.new_width / original_width
        f_y_new = f_y * self.new_height / original_height
        f_label_new = (f_x_new, f_y_new)

        od_x_new = od_x * self.new_width / original_width
        od_y_new = od_y * self.new_height / original_height
        od_label_new = (od_x_new, od_y_new)

        return image_new, (f_label_new, od_label_new)


class ResizedCenterCrop:
    """Resize the image and coordinate labels.
    First crops a square from the center, then resizes.
    """

    def __init__(self, new_size=(350, 350)):
        self.new_width = new_size[0]
        self.new_height = new_size[1]

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        f_x, f_y = f_label
        od_x, od_y = od_label
        original_width, original_height = image.size

        image = tf.center_crop(image, original_height)  # assume landscape
        f_x, f_y = f_x - (original_width - original_height) / 2, f_y
        od_x, od_y = od_x - (original_width - original_height) / 2, od_y
        original_width = original_height

        image = tf.resize(image, (self.new_height, self.new_width))
        f_x_new = f_x * self.new_width / original_width
        f_y_new = f_y * self.new_height / original_height
        f_label_new = (f_x_new, f_y_new)

        od_x_new = od_x * self.new_width / original_width
        od_y_new = od_y * self.new_height / original_height
        od_label_new = (od_x_new, od_y_new)

        return image, (f_label_new, od_label_new)


class ResizedCenterCropAndPad:
    """Resize the image and coordinate labels.
    First crops a square from the center, then resizes. Additionally pads black to top and bottom,
    if needed.
    """

    def __init__(self, new_size=(350, 350)):
        self.new_width = new_size[0]
        self.new_height = new_size[1]

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        f_x, f_y = f_label
        od_x, od_y = od_label

        if self._is_cut(image):
            pad = 50

            # Resize the image to a height of new_size + pad keeping the aspect ratio
            w, h = image.size
            ratio = w / h
            new_h = self.new_height + pad
            new_w = int(new_h * ratio)

            image = tf.resize(image, (new_h, new_w))

            f_x, f_y = f_x * new_w / w, f_y * new_h / h
            od_x, od_y = od_x * new_w / w, od_y * new_h / h

            assert (
                image.size[1] == new_h
            ), f"Image height should be {new_h}, but is {image.size[1]}"

            # Pad the image top and bottom
            image = tf.pad(
                image, (0, pad // 2, 0, pad // 2), padding_mode="constant", fill=0
            )
            f_y += pad // 2
            od_y += pad // 2

        f_label = (f_x, f_y)
        od_label = (od_x, od_y)

        image, (f_label, od_label) = ResizedCenterCrop(
            (self.new_width, self.new_height)
        )((image, (f_label, od_label)))

        return image, (f_label, od_label)

    def _is_cut(self, pil_image, thresh=0.5):
        # Center crop the image, then check if the top and bottom have enough black or are cut off
        image, (_, _) = ResizedCenterCrop((350, 350))((pil_image, ((0, 0), (0, 0))))
        image = np.array(image)

        top = image[:10, :, :]
        bottom = image[-10:, :, :]

        frac1 = np.mean(top == 0)
        frac2 = np.mean(bottom == 0)
        frac = (frac1 + frac2) / 2
        return frac < thresh


class RandomZoom:
    """Zoom a bit out of the image with probability p and adjust coordinate labels."""

    def __init__(self, p=0.5, max_zoom=0.25):
        if not 0 <= p <= 1:
            raise ValueError(
                f"Variable p is a probability, should be float between 0 to 1"
            )
        self.p = p  # float between 0 to 1 represents the probability of zooming
        if not -1 <= max_zoom <= 1:
            raise ValueError(f"Variable max_zoom should be float between -1 to 1")
        self.max_zoom = max_zoom

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        if np.random.random() < self.p:
            if self.max_zoom < 0:
                zoom_factor = 1 - np.random.uniform(0, -self.max_zoom)
            else:
                zoom_factor = 1 + np.random.uniform(0, self.max_zoom)
            image = tf.affine(
                image, translate=(0, 0), angle=0, scale=zoom_factor, shear=0
            )
            f_label = f_x * zoom_factor, f_y * zoom_factor
            od_label = od_x * zoom_factor, od_y * zoom_factor
            # Shouldn't it be center_x + (x - center_x) * zoom_factor?
        return image, (f_label, od_label)


class RandomHorizontalFlip:
    """Horizontal flip the image with probability p and adjust coordinate labels."""

    def __init__(self, p=0.5):
        if not 0 <= p <= 1:
            raise ValueError(
                f"Variable p is a probability, should be float between 0 to 1"
            )
        self.p = p  # float between 0 to 1 represents the probability of flipping

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        if np.random.random() < self.p:
            image = tf.hflip(image)
            f_label = w - f_x, f_y
            od_label = w - od_x, od_y
        return image, (f_label, od_label)


class RandomVerticalFlip:
    """Vertically flip the image with probability p and adjust coordinate labels."""

    def __init__(self, p=0.5):
        if not 0 <= p <= 1:
            raise ValueError(
                f"Variable p is a probability, should be float between 0 to 1"
            )
        self.p = p  # float between 0 to 1 represents the probability of flipping

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        if np.random.random() < self.p:
            image = tf.vflip(image)
            f_label = f_x, h - f_y
            od_label = od_x, h - od_y
        return image, (f_label, od_label)


class RandomTranslation:
    """Translate the image and adjust the coordinate labels accordingly."""

    def __init__(self, max_translation=(0.1, 0.1)):
        if (not 0 <= max_translation[0] <= 1) or (not 0 <= max_translation[1] <= 1):
            raise ValueError(f"Variable max_translation should be float between 0 to 1")
        self.max_translation_x = max_translation[0]
        self.max_translation_y = max_translation[1]

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        x_translate = int(
            np.random.uniform(-self.max_translation_x, self.max_translation_x) * w
        )
        y_translate = int(
            np.random.uniform(-self.max_translation_y, self.max_translation_y) * h
        )
        image = tf.affine(
            image, translate=(x_translate, y_translate), angle=0, scale=1, shear=0
        )
        f_label = f_x + x_translate, f_y + y_translate
        od_label = od_x + x_translate, od_y + y_translate
        return image, (f_label, od_label)


class RandomRotation:
    """Rotate the image and adjust the coordinate labels accordingly."""

    def __init__(self, max_angle=15):
        if not 0 <= max_angle <= 180:
            raise ValueError(f"Variable max_angle should be float between 0 to 180")
        self.max_angle = max_angle

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        angle = np.random.uniform(-self.max_angle, self.max_angle)
        image = tf.affine(image, translate=(0, 0), angle=angle, scale=1, shear=0)
        f_label = (f_x - w / 2) * np.cos(np.radians(angle)) - (f_y - h / 2) * np.sin(
            np.radians(angle)
        ) + w / 2, (f_x - w / 2) * np.sin(np.radians(angle)) + (f_y - h / 2) * np.cos(
            np.radians(angle)
        ) + h / 2
        od_label = (od_x - w / 2) * np.cos(np.radians(angle)) - (od_y - h / 2) * np.sin(
            np.radians(angle)
        ) + w / 2, (od_x - w / 2) * np.sin(np.radians(angle)) + (od_y - h / 2) * np.cos(
            np.radians(angle)
        ) + h / 2

        return image, (f_label, od_label)


class RandomSqueeze:
    """Squeeze the image inside a range of values in x and y dimensions and adjust the coordinate
    labels accordingly.
    """

    def __init__(self, max_squeeze=(0.1, 0.1)):
        if (not 0 <= max_squeeze[0] <= 1) or (not 0 <= max_squeeze[1] <= 1):
            raise ValueError(f"Variable max_squeeze should be float between 0 to 1")
        self.max_squeeze_x = max_squeeze[0]
        self.max_squeeze_y = max_squeeze[1]

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label
        x_squeeze = 1 + np.random.uniform(-self.max_squeeze_x, self.max_squeeze_x)
        y_squeeze = 1 + np.random.uniform(-self.max_squeeze_y, self.max_squeeze_y)
        image, (f_label, od_label) = Resize((int(w * x_squeeze), int(h * y_squeeze)))(
            image_label_sample
        )
        f_label = f_x * x_squeeze, f_y * y_squeeze
        od_label = od_x * x_squeeze, od_y * y_squeeze

        return image, (f_label, od_label)


class ImageAdjustment:
    """Change the brightness and contrast of the image and apply Gamma correction.
    No need to change the label."""

    def __init__(
        self, p=0.5, brightness_factor=0.8, contrast_factor=0.8, gamma_factor=0.4
    ):
        if not 0 <= p <= 1:
            raise ValueError(
                f"Variable p is a probability, should be float between 0 to 1"
            )
        self.p = p
        self.brightness_factor = brightness_factor
        self.contrast_factor = contrast_factor
        self.gamma_factor = gamma_factor

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]

        if np.random.random() < self.p:
            brightness_factor = 1 + np.random.uniform(
                -self.brightness_factor, self.brightness_factor
            )
            image = tf.adjust_brightness(image, brightness_factor)

        if np.random.random() < self.p:
            contrast_factor = 1 + np.random.uniform(
                -self.brightness_factor, self.brightness_factor
            )
            image = tf.adjust_contrast(image, contrast_factor)

        if np.random.random() < self.p:
            gamma_factor = 1 + np.random.uniform(
                -self.brightness_factor, self.brightness_factor
            )
            image = tf.adjust_gamma(image, gamma_factor)

        return image, (f_label, od_label)


class ToTensor:
    """Convert image and labels to tensors."""

    def __init__(self, scale_labels=True):
        self.scale_labels = scale_labels

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        w, h = image.size
        f_x, f_y = f_label
        od_x, od_y = od_label

        image = tf.to_tensor(image)

        if self.scale_labels:
            f_label = f_x / w, f_y / h
            od_label = od_x / w, od_y / h
        f_label = torch.tensor(f_label, dtype=torch.float32)
        od_label = torch.tensor(od_label, dtype=torch.float32)

        return image, (f_label, od_label)


class ToPILImage:
    """Convert a tensor image to a PIL image and convert coordinate labels to float (-lists)."""

    def __init__(self, unscale_labels=True):
        self.unscale_labels = unscale_labels

    def __call__(self, image_label_sample):
        image = image_label_sample[0]
        f_label, od_label = image_label_sample[1]
        f_label, od_label = f_label.tolist(), od_label.tolist()

        image = tf.to_pil_image(image)
        w, h = image.size

        if self.unscale_labels:
            f_x, f_y = f_label
            od_x, od_y = od_label

            f_label = f_x * w, f_y * h
            od_label = od_x * w, od_y * h

        return image, (f_label, od_label)



# From https://github.com/ruc-aimc-lab/SuperRetina
"""@inproceedings{liu2022SuperRetina,
  title={Semi-Supervised Keypoint Detector and Descriptor for Retinal Image Matching},
  author={Jiazhen Liu and Xirong Li and Qijie Wei and Jie Xu and Dayong Ding},
  booktitle={Proceedings of the 17th European Conference on Computer Vision (ECCV)},
  year={2022}
}""" 
from .inference import register, enhance, DEFAULT_CONFIG, WEIGHT_PATH, SuperRetina
from .inference import get_config as get_registration_config
from .inference import load_model as load_registration_model


import random
import sys
import time

from torch.nn import functional as F
import torch
import torch.nn as nn

from ..loss import DiceBCELoss, DiceLoss
from ..loss import triplet_margin_loss_gor, triplet_margin_loss_gor_one, sos_reg

from ..common import remove_borders, sample_keypoint_desc, simple_nms, nms, \
    sample_descriptors
from ..common import get_gaussian_kernel, affine_images

from ..common import sample_keypoint_desc, nms, simple_nms


def update_value_map(value_map, points, value_map_config):
    """
    Update value maps used for recording learned keypoints from PKE,
    and getting the final learned keypoints which are combined of previous learned keypoints.
    :param value_map: previous value maps
    :param points: the learned keypoints in this epoch
    :param value_map_config:
    :return: the final learned keypoints combined of previous learning points
    """

    raw_value_map = value_map.clone()
    # used for record areas of value=0
    raw_value_map[value_map == 0] = -1
    area_set = value_map_config['area']
    area = area_set // 2

    value_increase_point = value_map_config['value_increase_point']
    value_increase_area = value_map_config['value_increase_area']
    value_decay = value_map_config['value_decay']

    h, w = value_map[0].shape
    for (x, y) in points:
        y_d = y - area // 2 if y - area // 2 > 0 else 0
        y_u = y + area // 2 if y + area // 2 < h else h
        x_l = x - area // 2 if x - area // 2 > 0 else 0
        x_r = x + area // 2 if x + area // 2 < w else w
        tmp = value_map[0, y_d:y_u, x_l:x_r]
        if value_map[0, y, x] != 0 or tmp.sum() == 0:
            value_map[0, y, x] += value_increase_point  # if there is no learned point before, then add a high value
        else:
            tmp[tmp > 0] += value_increase_area
            value_map[0, y_d:y_u, x_l:x_r] = tmp

    value_map[torch.where(
        value_map == raw_value_map)] -= value_decay  # value decay of positions that don't appear this time

    tmp = value_map.detach().clone()

    tmp = simple_nms(tmp.unsqueeze(0).float(), area_set*2)
    tmp = tmp.squeeze()

    final_points = torch.nonzero(tmp >= value_increase_point)
    final_points = torch.flip(final_points, [1]).long()  # to x, y
    return final_points


def mapping_points(grid, points, h, w):
    """ Using grid_inverse to apply affine transform on geo_points
        :return point set and its corresponding affine point set
    """

    grid_points = [(grid[s, k[:, 1].long(), k[:, 0].long()]) for s, k in
                   enumerate(points)]
    filter_points = []
    affine_points = []
    for s, k in enumerate(grid_points):  # filter bad geo_points
        idx = (k[:, 0] < 1) & (k[:, 0] > -1) & (k[:, 1] < 1) & (
                k[:, 1] > -1)
        gp = grid_points[s][idx]
        gp[:, 0] = (gp[:, 0] + 1) / 2 * (w - 1)
        gp[:, 1] = (gp[:, 1] + 1) / 2 * (h - 1)
        affine_points.append(gp)
        filter_points.append(points[s][idx])

    return filter_points, affine_points


def content_filter(descriptor_pred, affine_descriptor_pred, geo_points,
                   affine_geo_points, content_thresh=0.7, scale=8):
    """
    content-based matching in paper
    :param descriptor_pred: descriptors of input_image images
    :param affine_descriptor_pred: descriptors of affine images
    :param geo_points: 
    :param affine_geo_points:
    :param content_thresh:
    :param scale: down sampling size of descriptor_pred
    :return: content-filtered keypoints
    """

    descriptors = [sample_keypoint_desc(k[None], d[None], scale)[0].permute(1, 0)
                   for k, d in zip(geo_points, descriptor_pred)]
    aff_descriptors = [sample_keypoint_desc(k[None], d[None], scale)[0].permute(1, 0)
                       for k, d in zip(affine_geo_points, affine_descriptor_pred)]
    content_points = []
    affine_content_points = []
    dist = [torch.norm(descriptors[d][:, None] - aff_descriptors[d], dim=2, p=2)
            for d in range(len(descriptors))]
    for i in range(len(dist)):
        D = dist[i]
        if len(D) <= 1:
            content_points.append([])
            affine_content_points.append([])
            continue
        val, ind = torch.topk(D, 2, dim=1, largest=False)

        arange = torch.arange(len(D))
        # rule1 spatial correspondence
        c1 = ind[:, 0] == arange.to(ind.device)
        # rule2 pass the ratio test
        c2 = val[:, 0] < val[:, 1] * content_thresh

        check = c2 * c1
        content_points.append(geo_points[i][check])
        affine_content_points.append(affine_geo_points[i][check])
    return content_points, affine_content_points


def geometric_filter(affine_detector_pred, points, affine_points, max_num=1024, geometric_thresh=0.5):
    """
    geometric matching in paper
    :param affine_detector_pred: geo_points probability of affine image
    :param points: nms results of input_image image
    :param affine_points: nms results of affine image
    :param max_num: maximum number of learned keypoints
    :param geometric_thresh: 
    :return: geometric-filtered keypoints
    """
    geo_points = []
    affine_geo_points = []
    for s, k in enumerate(affine_points):
        sample_aff_values = affine_detector_pred[s, 0, k[:, 1].long(), k[:, 0].long()]
        check = sample_aff_values.squeeze() >= geometric_thresh
        geo_points.append(points[s][check][:max_num])
        affine_geo_points.append(k[check][:max_num])

    return geo_points, affine_geo_points


def pke_learn(detector_pred, descriptor_pred, grid_inverse, affine_detector_pred,
              affine_descriptor_pred, kernel, loss_cal, label_point_positions,
              value_map, config, PKE_learn=True):
    """
    pke process used for detector
    :param detector_pred: probability map from raw image
    :param descriptor_pred: prediction of descriptor_pred network
    :param kernel: used for gaussian heatmaps
    :param mask_kernel: used for masking initial keypoints
    :param grid_inverse: used for inverse
    :param loss_cal: loss (default is dice)
    :param label_point_positions: positions of keypoints on labels
    :param value_map: value map for recoding and selecting learned geo_points
    :param pke_learn: whether to use PKE
    :return: loss of detector, num of additional geo_points, updated value maps and enhanced labels
    """
    # used for masking initial keypoints on enhanced labels
    initial_label = F.conv2d(label_point_positions, kernel,
                             stride=1, padding=(kernel.shape[-1] - 1) // 2)
    initial_label[initial_label > 1] = 1

    if not PKE_learn:
        return loss_cal(detector_pred, initial_label.to(detector_pred)), 0, None, None, initial_label

    nms_size = config['nms_size']
    nms_thresh = config['nms_thresh']
    scale = 8

    enhanced_label = None
    geometric_thresh = config['geometric_thresh']
    content_thresh = config['content_thresh']
    with torch.no_grad():
        h, w = detector_pred.shape[2:]

        # number of learned points
        number_pts = 0
        points = nms(detector_pred, nms_thresh=nms_thresh, nms_size=nms_size,
                     detector_label=initial_label, mask=True)

        # geometric matching
        points, affine_points = mapping_points(grid_inverse, points, h, w)
        geo_points, affine_geo_points = geometric_filter(affine_detector_pred, points, affine_points,
                                                         geometric_thresh=geometric_thresh)


        # content matching
        content_points, affine_contend_points = content_filter(descriptor_pred, affine_descriptor_pred, geo_points,
                                                               affine_geo_points, content_thresh=content_thresh,
                                                               scale=scale)
        enhanced_label_pts = []
        for step in range(len(content_points)):
            # used to combine initial points and learned points
            positions = torch.where(label_point_positions[step, 0] == 1)
            if len(positions) == 2:
                positions = torch.cat((positions[1].unsqueeze(-1), positions[0].unsqueeze(-1)), -1)
            else:
                positions = positions[0]

            final_points = update_value_map(value_map[step], content_points[step], config)

            # final_points = torch.cat((final_points, positions))

            temp_label = torch.zeros([h, w]).to(detector_pred.device)

            temp_label[final_points[:, 1], final_points[:, 0]] = 0.5
            temp_label[positions[:, 1], positions[:, 0]] = 1

            enhanced_kps = nms(temp_label.unsqueeze(0).unsqueeze(0), 0.1, 10)[0]
            if len(enhanced_kps) < len(positions):
                enhanced_kps = positions
            # print(len(final_points), len(positions), len(enhanced_kps))
            number_pts += (len(enhanced_kps) - len(positions))
            # number_pts += (len(enhanced_kps) - len(positions)) if (len(enhanced_kps) - len(positions)) > 0 else 0

            temp_label[:] = 0
            temp_label[enhanced_kps[:, 1], enhanced_kps[:, 0]] = 1

            enhanced_label_pts.append(temp_label.unsqueeze(0).unsqueeze(0))

            temp_label = F.conv2d(temp_label.unsqueeze(0).unsqueeze(0), kernel, stride=1,
                                  padding=(kernel.shape[-1] - 1) // 2)  # generating gaussian heatmaps
            temp_label[temp_label > 1] = 1

            if enhanced_label is None:
                enhanced_label = temp_label
            else:
                enhanced_label = torch.cat((enhanced_label, temp_label))

    enhanced_label_pts = torch.cat(enhanced_label_pts)
    affine_pred_inverse = F.grid_sample(affine_detector_pred, grid_inverse, align_corners=True)

    loss1 = loss_cal(detector_pred, enhanced_label)  # L_geo
    loss2 = loss_cal(detector_pred, affine_pred_inverse)  # L_clf
    # pred_mask = (enhanced_label > 0) & (affine_pred_inverse != 0)
    # loss2 = loss_cal(detector_pred[pred_mask], affine_pred_inverse[pred_mask])  # L_clf

    # mask_pred = grid_inverse
    # loss2 = loss_cal(detector_pred[mask_pred], affine_pred_inverse[mask_pred])  # L_clf

    loss = loss1+loss2

    return loss, number_pts, value_map, enhanced_label_pts, enhanced_label


def double_conv(in_channels, out_channels):
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, 3, padding=1),
        nn.ReLU(inplace=True),
        nn.Conv2d(out_channels, out_channels, 3, padding=1),
        nn.ReLU(inplace=True)
    )


class SuperRetina(nn.Module):
    def __init__(self, config=None, device='cpu', n_class=1):
        super().__init__()

        self.PKE_learn = True
        self.relu = torch.nn.ReLU(inplace=True)
        self.pool = torch.nn.MaxPool2d(kernel_size=2, stride=2)
        c1, c2, c3, c4, c5, d1, d2 = 64, 64, 128, 128, 256, 256, 256
        # Shared Encoder.
        self.conv1a = torch.nn.Conv2d(1, c1, kernel_size=3, stride=1, padding=1)
        self.conv1b = torch.nn.Conv2d(c1, c1, kernel_size=3, stride=1, padding=1)

        self.conv2a = torch.nn.Conv2d(c1, c2, kernel_size=3, stride=1, padding=1)
        self.conv2b = torch.nn.Conv2d(c2, c2, kernel_size=3, stride=1, padding=1)

        self.conv3a = torch.nn.Conv2d(c2, c3, kernel_size=3, stride=1, padding=1)
        self.conv3b = torch.nn.Conv2d(c3, c3, kernel_size=3, stride=1, padding=1)

        self.conv4a = torch.nn.Conv2d(c3, c4, kernel_size=3, stride=1, padding=1)
        self.conv4b = torch.nn.Conv2d(c4, c4, kernel_size=3, stride=1, padding=1)

        # Descriptor Head.
        self.convDa = torch.nn.Conv2d(c4, c5, kernel_size=3, stride=1, padding=1)
        self.convDb = torch.nn.Conv2d(c5, d1, kernel_size=4, stride=2, padding=0)
        self.convDc = torch.nn.Conv2d(d1, d2, kernel_size=1, stride=1, padding=0)

        self.trans_conv = nn.ConvTranspose2d(d1, d2, 2, stride=2)

        # Detector Head
        self.upsample = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.dconv_up3 = double_conv(c3 + c4, c3)
        self.dconv_up2 = double_conv(c2 + c3, c2)
        self.dconv_up1 = double_conv(c1 + c2, c1)

        self.conv_last = nn.Conv2d(c1, n_class, kernel_size=1)

        if config is not None:
            self.config = config

            self.nms_size = config['nms_size']
            self.nms_thresh = config['nms_thresh']
            self.scale = 8

            self.dice = DiceLoss()

            self.kernel = get_gaussian_kernel(kernlen=config['gaussian_kernel_size'],
                                              nsig=config['gaussian_sigma']).to(device)

        self.to(device)

    def network(self, x):
        x = self.relu(self.conv1a(x))
        conv1 = self.relu(self.conv1b(x))
        x = self.pool(conv1)
        x = self.relu(self.conv2a(x))
        conv2 = self.relu(self.conv2b(x))
        x = self.pool(conv2)
        x = self.relu(self.conv3a(x))
        conv3 = self.relu(self.conv3b(x))
        x = self.pool(conv3)
        x = self.relu(self.conv4a(x))
        x = self.relu(self.conv4b(x))

        # Descriptor Head.
        cDa = self.relu(self.convDa(x))
        cDb = self.relu(self.convDb(cDa))
        desc = self.convDc(cDb)

        dn = torch.norm(desc, p=2, dim=1)  # Compute the norm.
        desc = desc.div(torch.unsqueeze(dn, 1))  # Divide by norm to normalize.

        desc = self.trans_conv(desc)

        cPa = self.upsample(x)
        cPa = torch.cat([cPa, conv3], dim=1)

        cPa = self.dconv_up3(cPa)
        cPa = self.upsample(cPa)
        cPa = torch.cat([cPa, conv2], dim=1)

        cPa = self.dconv_up2(cPa)
        cPa = self.upsample(cPa)
        cPa = torch.cat([cPa, conv1], dim=1)

        cPa = self.dconv_up1(cPa)

        semi = self.conv_last(cPa)
        semi = torch.sigmoid(semi)

        return semi, desc

    def descriptor_loss(self, detector_pred, label_point_positions, descriptor_pred,
                        affine_descriptor_pred, grid_inverse, affine_detector_pred=None):
        """
        calculate descriptor loss, construct triples on raw images and affine images
        :param detector_pred: output of detector network
        :param label_point_positions: initial label points
        :param descriptor_pred: output of descriptor network
        :param affine_descriptor_pred: output of descriptor network, with affine images as input
        :param grid_inverse: used for inverse affine transformation
        :return: descriptor loss (triplet loss)
        """

        # sample keypoints on initial labels
        # label_descriptors, label_affine_descriptors, label_keypoints = \
        #     sample_descriptors(label_point_positions, descriptor_pred, affine_descriptor_pred, grid_inverse,
        #                        nms_size=self.nms_size, nms_thresh=self.nms_thresh, scale=self.scale)
        #
        # for s, kps in enumerate(label_keypoints):
        #     label_mask = torch.zeros(detector_pred[s].shape).to(detector_pred)
        #     label_mask[0, kps[:, 1].long(), kps[:, 0].long()] = 1
        #     label_mask = F.conv2d(label_mask.unsqueeze(0), self.mask_kernel, stride=1,
        #                           padding=(self.mask_kernel.shape[-1] - 1) // 2)
        #     detector_pred[s][label_mask[0] > 1e-5] = 0
        if not self.PKE_learn:
            detector_pred[:] = 0  # only learn from the initial labels
        detector_pred[label_point_positions == 1] = 10
        descriptors, affine_descriptors, keypoints = \
            sample_descriptors(detector_pred, descriptor_pred, affine_descriptor_pred, grid_inverse,
                               nms_size=self.nms_size, nms_thresh=self.nms_thresh, scale=self.scale,
                               affine_detector_pred=affine_detector_pred)

        # descriptors_tmp = []
        # affine_descriptor_tmp = []
        # for i in range(len(descriptors)):
        #     descriptors_tmp.append(torch.cat((descriptors[i], label_descriptors[i]), -1))
        #     affine_descriptor_tmp.append(torch.cat((affine_descriptors[i], label_affine_descriptors[i]), -1))
        # descriptors = descriptors_tmp
        # affine_descriptors = affine_descriptor_tmp

        positive = []
        negatives_hard = []
        negatives_random = []
        anchor = []
        D = descriptor_pred.shape[1]
        for i in range(len(affine_descriptors)):
            if affine_descriptors[i].shape[1] == 0:
                continue
            descriptor = descriptors[i]
            affine_descriptor = affine_descriptors[i]

            n = affine_descriptors[i].shape[1]
            if n > 1000:  # avoid OOM
                return torch.tensor(0., requires_grad=True).to(descriptor_pred), False

            descriptor = descriptor.view(D, -1, 1)
            affine_descriptor = affine_descriptor.view(D, 1, -1)
            ar = torch.arange(n)

            # random
            neg_index2 = []
            if n == 1:
                neg_index2.append(0)
            else:
                for j in range(n):
                    t = j
                    while t == j:
                        t = random.randint(0, n - 1)
                    neg_index2.append(t)
            neg_index2 = torch.tensor(neg_index2, dtype=torch.long).to(affine_descriptor)

            # hard
            with torch.no_grad():
                dis = torch.norm(descriptor - affine_descriptor, dim=0)
                dis[ar, ar] = dis.max() + 1
                neg_index1 = dis.argmin(axis=1)

            positive.append(affine_descriptor[:, 0, :].permute(1, 0))
            anchor.append(descriptor[:, :, 0].permute(1, 0))
            negatives_hard.append(affine_descriptor[:, 0, neg_index1.long(), ].permute(1, 0))
            negatives_random.append(affine_descriptor[:, 0, neg_index2.long(), ].permute(1, 0))

        if len(positive) == 0:
            return torch.tensor(0., requires_grad=True).to(descriptor_pred), False

        positive = torch.cat(positive)
        anchor = torch.cat(anchor)
        negatives_hard = torch.cat(negatives_hard)
        negatives_random = torch.cat(negatives_random)

        positive = F.normalize(positive, dim=-1, p=2)
        anchor = F.normalize(anchor, dim=-1, p=2)
        negatives_hard = F.normalize(negatives_hard, dim=-1, p=2)
        negatives_random = F.normalize(negatives_random, dim=-1, p=2)

        loss = triplet_margin_loss_gor(anchor, positive, negatives_hard, negatives_random, margin=0.8)

        # can also add sos reg term .
        # reg_term = sos_reg(anchor, positive, KNN=True, k=1, eps=1e-8)
        # if not torch.isnan(reg_term) and reg_term > 0:
        #     loss = loss + 0.1 * reg_term

        return loss, True

    def forward(self, x, label_point_positions=None, value_map=None, learn_index=None):
        """
        In interface phase, only need to input x
        :param x: retinal images
        :param label_point_positions: positions of keypoints on labels
        :param value_map: value maps, used to record history learned geo_points
        :param learn_index: index of input data with detector labels
        :param phase: distinguish dataset
        :return: if training, return loss, else return predictions
        """

        detector_pred, descriptor_pred = self.network(x)
        enhanced_label_pts = None
        enhanced_label = None

        if label_point_positions is not None:
            if self.PKE_learn:
                loss_detector_num = len(learn_index[0])
                loss_descriptor_num = x.shape[0]
            else:
                loss_detector_num = len(learn_index[0])
                loss_descriptor_num = loss_detector_num

            number_pts = 0  # number of learned keypoints
            value_map_update = None
            loss_detector = torch.tensor(0., requires_grad=True).to(x)
            loss_descriptor = torch.tensor(0., requires_grad=True).to(x)

            with torch.no_grad():
                affine_x, grid, grid_inverse = affine_images(x, used_for='detector')
                affine_detector_pred, affine_descriptor_pred = self.network(affine_x)
            loss_cal = self.dice
            if len(learn_index[0]) != 0:
                loss_detector, number_pts, value_map_update, enhanced_label_pts, enhanced_label = \
                    pke_learn(detector_pred[learn_index], descriptor_pred[learn_index],
                              grid_inverse[learn_index], affine_detector_pred[learn_index],
                              affine_descriptor_pred[learn_index], self.kernel, loss_cal,
                              label_point_positions[learn_index], value_map[learn_index],
                              self.config, self.PKE_learn)

            #  For showing PKE process
            if enhanced_label_pts is not None:
                enhanced_label_pts_tmp = label_point_positions.clone()
                enhanced_label_pts_tmp[learn_index] = enhanced_label_pts
                enhanced_label_pts = enhanced_label_pts_tmp
            if enhanced_label is not None:
                enhanced_label_tmp = label_point_positions.clone()
                enhanced_label_tmp[learn_index] = enhanced_label
                enhanced_label = enhanced_label_tmp

            detector_pred_copy = detector_pred.clone().detach()
            # if value_map_update is not None:
            #     # optimize descriptors of recorded points
            #     detector_pred_copy[learn_index][value_map_update >=
            #                                     self.config['VALUE MAP'].getfloat('value_increase_point')] = 1
            #
            affine_x_for_desc, grid_for_desc, grid_inverse_for_desc = affine_images(x, used_for='descriptor')
            _, affine_descriptor_pred_for_desc = self.network(affine_x_for_desc)
            loss_descriptor, descriptor_train_flag = self.descriptor_loss(detector_pred_copy, label_point_positions,
                                                                          descriptor_pred,
                                                                          affine_descriptor_pred_for_desc,
                                                                          grid_inverse_for_desc)

            if self.PKE_learn and len(learn_index[0]) != 0:
                value_map[learn_index] = value_map_update
            loss = loss_detector + loss_descriptor

            return loss, number_pts, loss_detector.cpu().data.sum(), \
                   loss_descriptor.cpu().data.sum(), enhanced_label_pts, \
                   enhanced_label, detector_pred, loss_detector_num, loss_descriptor_num

        return detector_pred, descriptor_pred




import torch
from PIL import Image
from torchvision.transforms.functional import to_tensor, to_pil_image
import numpy as np
from typing import Union

class ImageTorchUtils:
    """Class for image manipulation based on Pytorch.
    
    Usage:
    ```python
    import ImageTorchUtils as Img
    # Use for single image manipulation
    # Pass a path to an image or an image as a numpy array, PIL Image or torch tensor
    # and chain the desired methods
    img_path = "path/to/image.jpg"
    img = Img(img_path).to_tensor().img # > torch.Tensor
    img = Img(img_path).to_tensor().to_cspace(from_cspace="rgb", to_cspace="bgr").img # > torch.Tensor in BGR

    # Alternatively, use it to get a batch of images from a list of images or a list of 
    # paths to images
    img_paths = ["path/to/image1.jpg", "path/to/image2.jpg"]
    img_batch = Img(img_paths).to_image_batch().img
    ```
    """
    def __init__(self, image: Union[str, Image.Image, np.ndarray, torch.Tensor, list]):
        self.img = image

    def to_tensor(self, from_cspace: str = 'rgb', to_cspace:str = 'rgb', silent: bool = False):
        """Converts the instance's image to a torch tensor of shape (C, H, W).

        Args:
            from_cspace (str, optional): Colorspace of the image. Defaults to 'rgb'.
            to_cspace (str, optional): Colorspace to convert the image to. Defaults to 'rgb'.
        Returns:
            torch.Tensor: Image as a torch tensor.
        """

        if self.is_batch():
            if None not in [from_cspace, to_cspace] and not silent:
                print("inside to_tensor, from_cspace and to_cspace are ignored for batch images. Use to_cspace instead.")
            return self

        if self.is_batch_like():
            if None not in [from_cspace, to_cspace] and not silent:
                print("inside to_tensor, from_cspace and to_cspace are ignored for batch images. Use to_cspace instead.")
            imgs = [ImageTorchUtils(img).to_tensor().img for img in self.img]
            self.img = torch.stack(imgs)
            return self
        
        from_cspace = from_cspace.lower().replace(' ', '')
        from_cspace = "gray" if from_cspace in ['gray', 'grey'] else from_cspace
        if from_cspace not in ['rgb', 'bgr', 'gray', 'grey']:
            raise ValueError("from_cspace should be one of ['rgb', 'bgr', 'gray'] but is", from_cspace)
        
        if isinstance(self.img, str):
            # PIL Image with shape (H, W, C) -> (C, H, W)
            self.img = to_tensor(Image.open(self.img)) 
            from_cspace = 'rgb'
        elif isinstance(self.img, Image.Image):
            # PIL Image with shape (H, W, C) -> (C, H, W)
            self.img = to_tensor(self.img)
        elif isinstance(self.img, np.ndarray):
            if len(self.img.shape) == 2:
                # Grayscale numpy array with shape (H, W) -> (1, H, W)
                self.img = np.expand_dims(self.img, 2)
                self.img = to_tensor(self.img)
            elif len(self.img.shape) == 3 and self.img.shape[0] == 1:
                # Grayscale numpy array with shape (1, H, W) -> (1, H, W)
                self.img = to_tensor(self.img)
            else:
                # RGB/BGR numpy array with shape (H, W, C) -> (C, H, W)
                self.img = to_tensor(Image.fromarray(self.img))
        elif isinstance(self.img, torch.Tensor):
            self.img = self.img
        else:
            raise ValueError("Image should be a string, numpy array (as from plt.imread or cv2.imread), PIL Image or torch tensor but is of type", type(self.img))
        
        self.img = self.to_cspace(from_cspace, to_cspace).img

        return self
    
    def squeeze(self, dim: int = 0):
        """Squeezes the instance's image tensor along the specified dimension.

        Args:
            dim (int, optional): Dimension to squeeze. Defaults to 0.

        Returns:
            torch.Tensor: Image tensor with the specified dimension squeezed.
        """
        if self.is_batch_like():
            # Recursively squeeze each image in the batch
            imgs = [ImageTorchUtils(img).squeeze(dim).img for img in self.img]
            return self

        if self.is_batch():
            # Recursively squeeze each image in the batch
            imgs = [ImageTorchUtils(img).squeeze(dim).img for img in self.img]
            self.img = torch.stack(imgs)
            return self
        
        if isinstance(self.img, torch.Tensor) and len(self.img.shape) == 4 and len(self.shape[dim]) == 1:
            self.img = self.img.squeeze(dim)
        elif isinstance(self.img, np.ndarray) and len(self.img.shape) == 4 and len(self.img.shape[dim]) == 1:
            self.img = np.squeeze(self.img, dim)
        elif isinstance(self.img, Image.Image) and len(np.array(self.img).shape) == 4 and len(np.array(self.img).shape[dim]) == 1:
            self.img = np.array(self.img).squeeze(dim)
            self.img = Image.fromarray(self.img)
        else:
            if not isinstance(self.img, (torch.Tensor, np.ndarray, Image.Image)):
                raise ValueError("Image should be a torch tensor, Image.Image or numpy array but is of type", type(self.img))        
        
        return self
    
    def to_cspace(self, from_cspace: str, to_cspace: str):
        """Converts the instance's image tensor from one colorspace to another.

        Args:
            from_cspace (str): Colorspace of the image.
            to_cspace (str): Colorspace to convert the image to.

        Returns:
            torch.Tensor: Image in the new colorspace.
        """
        if not isinstance(self.img, torch.Tensor):
            raise ValueError("Image should be a torch tensor but is of type", type(self.img))

        if self.is_batch():
            # Recursively convert each image in the batch
            imgs = [ImageTorchUtils(img).to_cspace(from_cspace, to_cspace).img for img in self.img]
            self.img = torch.stack(imgs)
            return self

        from_cspace = from_cspace.lower().replace(' ', '')
        to_cspace = to_cspace.lower().replace(' ', '')
        from_cspace = "gray" if from_cspace in ['gray', 'grey'] else from_cspace
        to_cspace = "gray" if to_cspace in ['gray', 'grey'] else to_cspace
        if from_cspace not in ['rgb', 'bgr', 'gray', 'grey']:
            raise ValueError("from_cspace should be one of ['rgb', 'bgr', 'gray'] but is", from_cspace)
        if to_cspace not in ['rgb', 'bgr', 'gray', 'grey']:
            raise ValueError("to_cspace should be one of ['rgb', 'bgr', 'gray'] but is", to_cspace)
        
        if from_cspace == 'gray':
            self.img = self.img.expand(3, -1, -1)      
        elif from_cspace == 'bgr':
            self.img = self.img.flip(0)
        elif from_cspace == 'rgb':
            pass
            
        if to_cspace == 'gray':
            self.img = self.img.mean(0, keepdim=True)
        elif to_cspace == 'bgr':
            self.img = self.img.flip(0)
        elif to_cspace == 'rgb':
            pass
        return self
    
    def set_channel_dim(self, channel_dim: int):
        """Sets the channel dimension of the instance's image tensor.

        Args:
            dim (int): The desired channel dimension.
        """       
        if self.is_batch_like():
            # Recursively set the channel dimension for each image in the batch
            input_type = type(self.img[0])
            imgs = [ImageTorchUtils(img).set_channel_dim(channel_dim).img for img in self.img]
            if input_type == np.ndarray:
                imgs = np.array([np.array(img) for img in imgs])
            else:
                self.img = torch.stack(imgs)
            return self

        input_type = type(self.img)

        if not isinstance(self.img, torch.Tensor):
            self.img = self.to_tensor().img

        if len(self.img.shape) != 3:
            raise ValueError("Image should have 3 dimensions but has", len(self.img.shape))
        
        # Index of the current channel dimension: Smallest shape
        current_channel_dim = self.img.shape.index(min(self.img.shape))

        if channel_dim == -1:
            channel_dim = len(self.img.shape) - 1
        if channel_dim >= len(self.img.shape):
            raise ValueError("channel_dim should be less than the number of dimensions of the image tensor but is", channel_dim)
        if channel_dim < 0:
            raise ValueError("channel_dim should be -1 or greater than or equal to 0 but is", channel_dim)

        if current_channel_dim != channel_dim:
            newdims = list(range(len(self.img.shape)))
            newdims.remove(current_channel_dim)
            newdims.insert(channel_dim, current_channel_dim)
            self.img = self.img.permute(*newdims)

        if input_type == np.ndarray:
            self.img = self.img.numpy()
        
        return self

    def to_pil(self):
        """Converts the instance's image tensor to a PIL Image.

        Returns:
            Image: The image as a PIL Image.
        """
        if not isinstance(self.img, torch.Tensor):
            raise ValueError("Image should be a torch tensor but is of type", type(self.img))
        
        if self.is_batch():
            # Recursively convert each image in the batch. Returning a list of PIL Images.
            imgs = [ImageTorchUtils(img).to_pil().img for img in self.img]
            self.img = imgs
            return self
        
        self.img = to_pil_image(self.img)
        return self
    
    def to_numpy(self, dtype: str="uint8"):
        """Converts the instance's image tensor to a numpy array (H, W, C).

        Args:
            image (torch.Tensor): Image tensor.
            dtype (str, optional): Data type of the numpy array. Defaults to "uint8".

        Returns:
            np.ndarray: Numpy array.
        """
        if isinstance(self.img, np.ndarray) and self.img.dtype == dtype:
            return self
        
        if self.is_batch_like():
            # Recursively convert each image in the batch. Returning an narray of numpy arrays.
            imgs = [ImageTorchUtils(img).to_numpy(dtype).img for img in self.img]
            self.img = np.array(imgs)
            return self
        
        # if not isinstance(self.img, torch.Tensor):
        # (H, W, C) as is standard for numpy arrays
        self.img = self.to_tensor().set_channel_dim(-1).img 

        if dtype == "uint8":
            if self.img.max() <= 1:
                self.img = (self.img * 255).byte()
            else:
                self.img = self.img.byte()
        elif dtype == "float32":
            self.img = self.img.float()
        else:
            raise ValueError("dtype should be one of ['uint8', 'float32'] but is", dtype)
        
        self.img = self.img.numpy()
        return self
    
    def to_uint8(self):
        """Converts the instance's image tensor to a numpy array with dtype uint8.

        Returns:
            np.ndarray: The image as a numpy array with dtype uint8.
        """
        if isinstance(self.img, np.ndarray) and self.img.dtype == np.uint8:
            return self

        if self.is_batch_like():
            # Recursively convert each image in the batch
            imgs = [ImageTorchUtils(img).to_uint8().img for img in self.img]
            self.img = np.array(imgs)
            return self

        if not isinstance(self.img, np.ndarray):
            self.img = self.to_numpy().img
        
        if self.img.dtype == np.float32:
            self.img = (self.img * 255).astype(np.uint8)
        return self
    
    def to_float32(self):
        """Converts the instance's image tensor to a numpy array with dtype float32.

        Returns:
            np.ndarray: The image as a numpy array with dtype float32.
        """
        if isinstance(self.img, np.ndarray) and self.img.dtype == np.float32:  
            return self

        if self.is_batch_like():
            # Recursively convert each image in the batch
            imgs = [ImageTorchUtils(img).to_float32().img for img in self.img]
            self.img = np.array(imgs)
            return self

        if not isinstance(self.img, np.ndarray):
            self.img = self.to_numpy().img
        
        if self.img.dtype == np.uint8:
            self.img = self.img.astype(np.float32) / 255
        return self
    
    def to_batch(self):
        """Converts the instance's image to a batch of images or the instance's
        set of images to a batch of image tensors. The instance's image can be
        a list, string, numpy array, PIL Image or torch tensor corresponding to
        a single image or a list of the same or a path or list of paths to images.
        
        Returns:
            torch.Tensor: Batch of images.
        """
        if self.is_batch():
            return self
        
        # Single instance to list
        if not self.is_batch_like():
            self.img = [self.img]
        
        # Paths, nd.arrays, PIL Images to image tensors
        if isinstance(self.img[0], (str, Image.Image, np.ndarray, torch.Tensor)):
            self.img = [ImageTorchUtils(image).to_tensor().img for image in self.img]
        else:
            raise ValueError("Image should be a list of paths, numpy arrays, PIL Images or torch tensors but is of type", type(self.img[0]))
        
        self.img = torch.stack(self.img)
        return self

    def is_batch(self) -> bool:
        """Checks if the instance's image is a tensor batch of tensor images.
        Use to_batch to convert lists or arrays of images or a single image to a batch of images.

        Returns:
            bool: True if the image is a batch of images, False otherwise.
        """
        return isinstance(self.img, torch.Tensor) and len(self.img.shape) == 4 and [isinstance(img, torch.Tensor) for img in self.img]
    
    def is_batch_like(self) -> bool:
        """Checks if the instance's image is a list or array of images or a single image.

        Returns:
            bool: True if the image is a list or array of images, False otherwise.
        """
        
        return self.is_batch() or isinstance(self.img, (list, tuple)) or len(np.array(self.img).shape) == 4


from types import SimpleNamespace
from typing import Union
import yaml
from pathlib import Path
import requests

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import torch
from torchvision import models
from torchvision.transforms.functional import to_pil_image, to_tensor
import matplotlib.pyplot as plt
import torch.nn as nn
from tqdm import tqdm
from datetime import datetime
import os
from PIL import ImageDraw, Image
import numpy as np
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    auc,
    roc_curve,
    precision_recall_curve,
)
from fundus_image_toolbox.utilities import ImageTorchUtils as Img
from fundus_image_toolbox.utilities import seed_everything

from .transforms import get_unnormalization, get_transforms
from .default import ENSEMBLE_MODELS, MODELS_DIR


def wget(link, target):
    # platform independent wget alternative
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.141 Safari/537.36"
    }
    r = requests.get(link, headers=headers, stream=True)
    downloaded_file = open(target, "wb")

    for chunk in r.iter_content(chunk_size=8192):
        if chunk:
            downloaded_file.write(chunk)


def download_weights(url="https://zenodo.org/records/11174749/files/weights.tar.gz"):
    MODELS_DIR.mkdir(parents=True, exist_ok=True)
    print("Downloading weights...")
    target = (MODELS_DIR / "weights.tar.gz").__str__()
    wget(url, target)
    print("Extracting weights...")
    os.system(f'tar -xzf {MODELS_DIR / "weights.tar.gz"} -C {MODELS_DIR}')
    print("Removing tar file...")
    os.remove(target)
    print("Done")


def plot_quality(fundus, conf: float, label: int, threshold: float = 0.5):
    """Plots the fundus image with a colorbar indicating the confidence score.

    Args:
        fundus (str, np.ndarray): Fundus image path or image as numpy array
        conf (float): Confidence score
        label (int): Label (0: ungradable, 1: gradable)
        threshold (float, optional): Threshold for binary classification. Defaults to 0.5.

    Returns:
        None
    """
    # Image to array with (H, W, C) shape
    fundus = Img(fundus).to_tensor().squeeze().to_numpy().set_channel_dim(-1).img

    if Img(fundus).is_batch_like():
        raise ValueError("Pass a single image, not a batch.")

    fig = plt.figure(figsize=(4.3, 4))
    gs = gridspec.GridSpec(1, 2, width_ratios=[25, 1])

    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1])

    # Plot the fundus image
    ax1.imshow(fundus)
    ax1.axis("off")
    ax1.text(
        0,
        20,
        f'{"Ungradeable" if label == 0 else "Gradeable"}',
        fontsize=12,
        color="white",
        backgroundcolor="black",
    )

    # Create a colorbar
    cmap = mpl.cm.RdYlGn
    norm = mpl.colors.Normalize(vmin=0, vmax=1)
    cb1 = mpl.colorbar.ColorbarBase(ax2, cmap=cmap, norm=norm, orientation="vertical")

    # Draw a horizontal line at the threshold with legend
    ax2.axhline(y=threshold, color="k", linestyle="-")

    # Draw an X at the confidence score
    ax2.scatter(0.5, conf, color="black", marker="x", s=50)

    plt.show()


class FundusQualityModel:
    """Fundus Quality Model with one output logit for binary classification"""

    def __init__(self, config):
        if isinstance(config, dict):
            config = SimpleNamespace(**config)
        self.config = config
        self.model = self._get_model(self.config.model_type)
        self.best_epoch = 0
        self.best_loss = float("inf")
        self.best_acc = 0.0
        self.model_selection_metric = (
            self.config.model_selection_metric
            if hasattr(self.config, "model_selection_metric")
            else "accuracy"
        )

        self.was_trained = False
        seed_everything(self.config.seed, silent=True)

        metrics = ["acc", "auroc", "auprc"]
        self.performance = {"train": {}, "val": {}, "test": {}}
        for metric in metrics:
            self.performance["train"][metric] = np.nan
            self.performance["val"][metric] = np.nan
            self.performance["test"][metric] = np.nan

        self.model.to(self.config.device)

        MODELS_DIR.mkdir(parents=True, exist_ok=True)
        self.timestamp = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
        self.checkpoint_path = (
            MODELS_DIR / f"{self.timestamp}/{self.config.model_type}_best.pt"
        )

        self.loss_fn = nn.BCEWithLogitsLoss()
        self.optimizer = torch.optim.Adam(
            self.model.parameters(),
            lr=self.config.lr,
            weight_decay=self.config.weight_decay,
        )

    def load_checkpoint(self, dir: str):
        dir = Path(dir)
        if dir.suffix == ".pt":
            dir = dir.parent
        if dir.name.endswith("/"):
            dir = dir.parent
        if str(dir) in ENSEMBLE_MODELS and not dir.is_dir():
            download_weights()
        self.timestamp = dir.name
        self.checkpoint_path = dir / f"{self.config.model_type}_best.pt"

        self.model.load_state_dict(
            torch.load(
                self.checkpoint_path, map_location=self.config.device, weights_only=True
            )
        )
        self.model.to(self.config.device)

        print(f"Model loaded from {self.checkpoint_path.parent.name}")

    def train(self, train_dataloader, val_dataloader):
        self.checkpoint_path.parent.mkdir(parents=True, exist_ok=True)

        self.loss_tracker = {"train": [], "val": []}
        self.acc_tracker = {"train": [], "val": []}

        for epoch in range(self.config.epochs):
            pbar = tqdm(total=len(train_dataloader))
            pbar.set_description(f"Epoch: {epoch+1} Training...")

            train_loss, train_acc = self._train_val_step(
                train_dataloader, self.loss_fn, self.optimizer, pbar
            )
            self.loss_tracker["train"].append(train_loss)
            self.acc_tracker["train"].append(train_acc)

            pbar = tqdm(total=len(val_dataloader))
            pbar.set_description(f"Epoch: {epoch+1} Validating...")

            val_loss, val_acc = self._train_val_step(
                val_dataloader, self.loss_fn, pbar=pbar
            )
            self.loss_tracker["val"].append(val_loss)
            self.acc_tracker["val"].append(val_acc)

            pbar.set_description(
                f"Epoch: {epoch+1}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}, Train Acc: {train_acc:.4f}, Val Acc: {val_acc:.4f}"
            )

            if val_acc > self.best_acc:
                self.best_acc = val_acc
                self.best_epoch = epoch
                if "loss" not in self.model_selection_metric:
                    torch.save(self.model.state_dict(), self.checkpoint_path)

            if val_loss < self.best_loss:
                self.best_loss = val_loss
                self.best_epoch = epoch
                if "loss" in self.model_selection_metric:
                    torch.save(self.model.state_dict(), self.checkpoint_path)

        pbar.close()

        self.was_trained = True

        print(
            f"Best epoch: {self.best_epoch}, Best loss: {self.best_loss}, Best acc: {self.best_acc}\n"
        )
        # Save summary file and loss plot
        self._plot_loss(save=True)

    def _train_val_step(self, dataloader, loss_fn, optimizer=None, pbar=None):
        if optimizer is not None:
            self.model.train()
        else:
            self.model.eval()

        running_loss = 0.0
        running_acc = 0.0

        for image_batch, label_batch in dataloader:
            pbar.update(1)

            image_batch, label_batch = image_batch.to(
                self.config.device
            ), label_batch.to(self.config.device)
            label_batch = label_batch.unsqueeze(1).float()

            with torch.set_grad_enabled(optimizer is not None):
                output = self.model(image_batch)
                loss = loss_fn(output, label_batch)
                running_loss += loss.item()

            with torch.no_grad():
                preds = torch.sigmoid(output)
                preds = torch.where(preds > 0.5, 1, 0)
                running_acc += torch.mean((preds == label_batch).float()).item()

            if optimizer:
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        with torch.no_grad():
            epoch_loss = running_loss / len(dataloader)
            epoch_acc = running_acc / len(dataloader)

        return epoch_loss, epoch_acc

    def predict_from_dataloader(
        self, dataloader, threshold=None, device=None, load_best=True, numpy_cpu=True
    ):
        if device is not None:
            self.config.device = device

        if load_best:
            self.load_checkpoint(self.checkpoint_path)

        self.model.eval()

        all_preds = []
        all_labels = []

        for image_batch, label_batch in dataloader:
            image_batch, label_batch = image_batch.to(
                self.config.device
            ), label_batch.to(self.config.device)

            with torch.no_grad():
                output = self.model(image_batch)
                preds = torch.sigmoid(output)
                all_preds.extend(preds.cpu().numpy())
                all_labels.extend(label_batch.cpu().numpy())

            all_preds = torch.tensor(np.array(all_preds))
            all_labels = torch.tensor(all_labels).squeeze()
            assert all_preds.shape[0] == all_labels.shape[0]

            if threshold:
                all_preds = torch.where(all_preds >= threshold, 1, 0).float().squeeze()
            else:
                all_preds = all_preds.squeeze()

            if numpy_cpu:
                all_preds = all_preds.cpu().numpy()
                all_labels = all_labels.cpu().numpy()

            assert all_preds.shape[0] == all_labels.shape[0]

        return all_preds, all_labels

    def predict_from_image(
        self,
        image: Union[str, torch.Tensor, np.ndarray],
        threshold: float = 0.5,
        device: str = None,
        load_best: bool = True,
        numpy_cpu: bool = True,
    ):
        """Predicts from a single image and returns the prediction (logit or class) as a numpy array.

        Args:
            image (str, torch.Tensor or numpy.ndarray): Image to predict from
            threshold (float, optional): Threshold to convert logit to class. Defaults to 0.5.
                classes: 0 if logit < threshold "ungradable", 1 if logit >= threshold "gradable"
            device (str, optional): Device to use. Defaults to None.
            load_best (bool, optional): Load best checkpoint. Defaults to True.
            numpy_cpu (bool, optional): Return numpy array in cpu. Defaults to True.

        Returns:
            numpy.ndarray: Prediction
        """
        image = Img(image).to_tensor().img
        image = get_transforms(split="test")(to_pil_image(image)).unsqueeze(0)

        return self.predict_from_batch(
            image, threshold, device, load_best, numpy_cpu, transform=False
        )

    def predict_from_batch(
        self,
        image_batch: Union[torch.Tensor, np.ndarray],
        threshold: float = 0.5,
        device: str = None,
        load_best: bool = True,
        numpy_cpu: bool = True,
        transform=True,
    ):
        """Predicts from a batch of images and returns the predictions (logits or classes) as a numpy array.

        Args:
            image_batch (torch.Tensor or numpy.ndarray): Batch of images to predict from
            threshold (float, optional): Threshold to convert logits to classes. Defaults to 0.5.
                classes: 0 if logit < threshold "ungradable", 1 if logit >= threshold "gradable"
            device (str, optional): Device to use. If None, model's device is used. Defaults to None.
            load_best (bool, optional): Load best checkpoint. Defaults to True.
            numpy_cpu (bool, optional): Return numpy array in cpu. Defaults to True.

        Returns:
            numpy.ndarray: Predictions
        """
        image_batch = Img(image_batch).to_batch().img
        if transform:
            image_batch = [
                get_transforms(split="test")(to_pil_image(image))
                for image in image_batch
            ]
            image_batch = torch.stack(image_batch)

        if device is not None:
            self.config.device = device

        if load_best:
            self.load_checkpoint(self.checkpoint_path)

        self.model.eval()

        image_batch = image_batch.to(self.config.device)

        with torch.no_grad():
            output = self.model(image_batch)
            preds = torch.sigmoid(output)

            if threshold:
                preds = torch.where(preds >= threshold, 1, 0).float()  # .squeeze()
            else:
                preds = preds  # .squeeze()

            if numpy_cpu:
                preds = preds.cpu().numpy()

        return preds

    def evaluate(
        self,
        test_dataloader=None,
        val_dataloader=None,
        train_dataloader=None,
        threshold=0.5,
        best=True,
        plot_auc=False,
    ):
        if (
            sum(
                1
                for _ in filter(
                    None.__ne__, [test_dataloader, val_dataloader, train_dataloader]
                )
            )
            != 1
        ):
            raise ValueError("You should pass exactly one dataloader.")

        if val_dataloader:
            loader = "val"
            dataloader = val_dataloader
        elif test_dataloader:
            loader = "test"
            dataloader = test_dataloader
        elif train_dataloader:
            loader = "train"
            dataloader = train_dataloader

        all_preds, all_labels = self.predict_from_dataloader(
            dataloader, threshold=threshold, load_best=best, numpy_cpu=False
        )

        binary_preds = torch.where(all_preds > threshold, 1, 0).float().squeeze()

        acc = torch.mean((binary_preds == all_labels).float()).item()
        fpr, tpr, _ = roc_curve(all_labels, all_preds)
        auroc = auc(fpr, tpr)
        precision_, recall_, _ = precision_recall_curve(all_labels, all_preds)
        auprc = auc(recall_, precision_)

        if plot_auc:
            fig, ax = plt.subplots()
            ax.plot(fpr, tpr, label=f"Ensemble (AUROC = {auroc:.2f})")
            ax.plot(recall_, precision_, label=f"Ensemble (AUPRC = {auprc:.2f})")
            ax.legend()
            plt.show()

        self.performance[loader]["acc"] = acc
        self.performance[loader]["auroc"] = auroc
        self.performance[loader]["auprc"] = auprc

        return self.performance[loader]

    def save_summary(self, log=None):
        summary_path = (
            MODELS_DIR / f"{self.timestamp}/{self.config.model_type}_summary.txt"
        )
        config_path = MODELS_DIR / f"{self.timestamp}/config.yaml"

        if not config_path.is_file():
            with open(config_path, "w") as c:
                yaml.dump(vars(self.config), c)

        with open(summary_path, "a") as f:
            f.write(datetime.now().strftime("%Y-%m-%d %H-%M-%S") + "\n")
            if log is not None:
                if isinstance(log, str):
                    f.write(log + "\n")
                elif isinstance(log, dict):
                    for key, value in log.items():
                        f.write(f"{key}: {value}\n")
                else:
                    print("Cannot log what was passed of type", type(log))

            if self.was_trained:
                for key, value in vars(self.config).items():
                    f.write(f"{key}: {value}\n")

                f.write(f"\nBest epoch: {self.best_epoch}\n")
                f.write(f"Best val loss: {self.best_loss}\n")
                f.write(f"Best val acc: {self.best_acc}\n")

            for loader in ["train", "val", "test"]:
                if not np.isnan(self.performance[loader]["acc"]):
                    f.write(f"{loader} performance:\n")
                    for metric in ["acc", "auroc", "auprc"]:
                        f.write(f"{metric}: {self.performance[loader][metric]:.4f}\n")
                    f.write("\n")

    def _plot_loss(self, save=False):
        fig, ax = plt.subplots(1, 2, figsize=(15, 5))
        for key in self.loss_tracker.keys():
            ax[0].plot(self.loss_tracker[key], label=key)
            ax[1].plot(self.acc_tracker[key], label=key)
        ax[0].set_title("Loss")
        ax[0].legend()
        ax[1].set_title("Accuracy")
        ax[1].legend()

        if save:
            plt.savefig(
                MODELS_DIR / f"{self.timestamp}/{self.config.model_type}_loss_plot.png"
            )

        plt.show()

    def plot_grid(self, dataloader, size=10, cols=6):
        self.model.eval()
        samples = np.random.choice(len(dataloader.dataset), size, replace=False)
        rows = np.ceil(len(samples) / cols).astype(int)

        fig, axs = plt.subplots(rows, cols, figsize=(cols * 2.5, rows * 2.5))
        axs = axs.flatten()

        size = len(axs) if size < len(axs) else size

        for i, idx in enumerate(samples):
            if i < size:
                image, label = dataloader.dataset[idx]
                image = image.unsqueeze(0).to(self.config.device)
                output = torch.sigmoid(self.model(image)).item()
                output = 0 if output < 0.5 else 1
                correct = output == label

                # Unnormalize
                image = get_unnormalization()(image.squeeze().cpu())
                image = image.permute(1, 2, 0)

                # Clip
                image = np.clip(image, 0, 1).numpy()

                # Rectangle
                col = "green" if label else "red"
                image = np.array(image * 255, dtype=np.uint8)
                image = Image.fromarray(image)
                draw = ImageDraw.Draw(image)
                img_size = image.size
                draw.rectangle([0, 0, img_size[0], img_size[1]], outline=col, width=10)
                image = np.array(image)

                axs[i].imshow(image)

                if not correct:
                    axs[i].set_title(f"wrong", color="red")

                axs[i].axis("off")

            else:
                axs[i].axis("off")
        plt.show()

    def _get_model(self, type, n_outs=1):
        if type == "resnet18":
            model = models.resnet18(weights="IMAGENET1K_V1")
            model.fc = torch.nn.Linear(512, n_outs)
        elif type == "resnet34":
            model = models.resnet34(weights="IMAGENET1K_V1")
            model.fc = torch.nn.Linear(512, n_outs)
        elif type == "resnet50":
            model = models.resnet50(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, n_outs)
        elif type == "resnet101":
            model = models.resnet101(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, n_outs)
        elif type == "resnet152":
            model = models.resnet152(weights="IMAGENET1K_V2")
            model.fc = torch.nn.Linear(2048, n_outs)
        elif type == "efficientnet-b0":
            model = models.efficientnet_b0(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1280, n_outs)
        elif type == "efficientnet-b1":
            model = models.efficientnet_b1(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1280, n_outs)
        elif type == "efficientnet-b2":
            model = models.efficientnet_b2(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1408, n_outs)
        elif type == "efficientnet-b3":
            model = models.efficientnet_b3(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1536, n_outs)
        elif type == "efficientnet-b4":
            model = models.efficientnet_b4(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(1792, n_outs)
        elif type == "efficientnet-b5":
            model = models.efficientnet_b5(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2048, n_outs)
        elif type == "efficientnet-b6":
            model = models.efficientnet_b6(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2304, n_outs)
        elif type == "efficientnet-b7":
            model = models.efficientnet_b7(weights="IMAGENET1K_V1")
            model.classifier = torch.nn.Linear(2560, n_outs)

        else:
            raise ValueError("Model type not supported")

        return model



import os


def set_seed(seed: int = 12345, silent=False) -> None:
    """Set seed for reproducibility."""
    try:
        import numpy as np

        np.random.seed(seed)
        if not silent:
            print(f"Numpy random seed was set as {seed}")
    except ImportError:
        print("Could not set seed for numpy: Numpy not found.")

    try:
        import torch

        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        if not silent:
            print(f"Torch random seed was set as {seed}")
    except ImportError:
        print("Could not set seed for torch: Torch not found.")

    try:
        import random

        random.seed(seed)
        if not silent:
            print(f"Random seed was set as {seed}")
    except ImportError:
        if not silent:
            print("Could not set seed for random: Random not found.")
        else:
            print()

    # Set a fixed value for the hash seed
    os.environ["PYTHONHASHSEED"] = str(seed)
    if not silent:
        print(f"PYTHONHASHSEED was set as {seed}")
        print()



import os
from setuptools import setup, find_packages

THIS_REPO = "https://github.com/berenslab/fundus_image_toolbox"
ROOT = os.path.join(os.path.dirname(__file__))

exclude_strings = ["test", "egg", "0_", "1_", "outdated"]
SUBMODULES = [
    f
    for f in os.listdir(ROOT)
    if os.path.isdir(f)
    and not f.startswith(".")
    and not f.startswith("__")
    and not any([exclude_string in f for exclude_string in exclude_strings])
]

with open(os.path.join(ROOT, "Readme.md"), "r", encoding="utf8") as f:
    long_description = f.read()

with open(os.path.join(ROOT, "requirements.txt")) as f:
    required = f.read().splitlines()

# Add subfolders as dependencies
# TODO remove temporary branch
branch_name = "main"
submodules_required = []
# [
#     f"{submodule} @ git+{THIS_REPO}@{branch_name}#egg={submodule}&subdirectory={submodule}"
#     for submodule in SUBMODULES
# ]

# Include all subpackages.
# - as pip dependencies: use install_requires=required+submodules_required, packages=[], package_dir={}
# - from local subdirectories: use find_packages() and to prevent extra import
#   hierarchy, add an __init__.py into every outer subpackage folder and import the functions there.
#   Caveat: All python files in the outer subpackage folder will be imported, too, such as setup.py,
#   which is why we don't do this here.
setup(
    name="fundus_image_toolbox",
    version="0.0.1",
    author="Julius Gervelmeyer et al.",
    author_email="Julius.Gervelmeyer@uni-tuebingen.de",
    description=long_description.split("\n")[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=THIS_REPO,
    packages=SUBMODULES,  # find_packages(),
    install_requires=required + submodules_required,
    # package_dir={"": ROOT},
    python_requires=">=3.9.0, <3.10",
    include_package_data=True,
)



from typing import List
from tqdm import tqdm
import torch

def get_pixel_mean_and_sd(loaders:List[torch.utils.data.DataLoader], device:str = "cuda:0"):
    """Compute the mean and standard deviation of a dataset.
    
    Args:
        loaders (list): list of torch data loaders; Use no augmentations here!
        device (str): torch device to use
    
    Returns:
        list: mean of the dataset for each channel
        list: standard deviation of the dataset for each channel
    """
    
    cnt = 0
    fst_moment = torch.empty(3).to(device)
    snd_moment = torch.empty(3).to(device)
    for loader in loaders:
        for (data, label) in tqdm(loader):
            b, c, h, w = data.shape
            nb_pixels = b * h * w
            sum_ = torch.sum(data, dim=[0, 2, 3])
            sum_of_square = torch.sum(data ** 2,
                                    dim=[0, 2, 3])
            fst_moment = (cnt * fst_moment + sum_) / (cnt + nb_pixels)
            snd_moment = (cnt * snd_moment + sum_of_square) / (cnt + nb_pixels)
            cnt += nb_pixels

    mean, std = fst_moment, torch.sqrt(snd_moment - fst_moment ** 2)        
    return mean, std


import math
import os
import random
import shutil
import time

import torch
from torchvision import transforms as T
from torch.nn import functional as F
import numpy as np
import cv2
import scipy.stats as st
from tqdm import tqdm
import matplotlib.pyplot as plt


def affine_images(images, used_for='detector'):
    """
    Perform affine transformation on images
    :param images: (B, C, H, W)
    :param keypoint_labels: corresponding labels
    :param value_map: value maps, used to record history learned geo_points
    :return: results of affine images, affine labels, affine value maps, affine transformed grid_inverse, inverse transformed grid_inverse
    """

    h, w = images.shape[2:]
    theta = None
    thetaI = None

    for i in range(len(images)):
        if used_for == 'detector':
            affine_params = T.RandomAffine(20).get_params(degrees=[-15, 15], translate=[0.2, 0.2],
                                                          scale_ranges=[0.9, 1.35],
                                                          shears=None, img_size=[h, w])
        else:
            affine_params = T.RandomAffine(20).get_params(degrees=[-3, 3], translate=[0.1, 0.1],
                                                          scale_ranges=[0.9, 1.1],
                                                          shears=None, img_size=[h, w])
        angle = -affine_params[0] * math.pi / 180
        theta_ = torch.tensor([
            [1 / affine_params[2] * math.cos(angle), math.sin(-angle), -affine_params[1][0] / images.shape[2]],
            [math.sin(angle), 1 / affine_params[2] * math.cos(angle), -affine_params[1][1] / images.shape[3]],
            [0, 0, 1]
        ], dtype=torch.float).to(images)
        thetaI_ = theta_.inverse()
        theta_ = theta_[:2]
        thetaI_ = thetaI_[:2]

        theta_ = theta_.unsqueeze(0)
        thetaI_ = thetaI_.unsqueeze(0)
        if theta is None:
            theta = theta_
        else:
            theta = torch.cat((theta, theta_))
        if thetaI is None:
            thetaI = thetaI_
        else:
            thetaI = torch.cat((thetaI, thetaI_))

    grid = F.affine_grid(theta, images.size(), align_corners=True)
    grid = grid.to(images)
    grid_inverse = F.affine_grid(thetaI, images.size(), align_corners=True)
    grid_inverse = grid_inverse.to(images)
    output = F.grid_sample(images, grid, align_corners=True)

    if used_for == 'descriptor':
        if random.random() >= 0.4:
            output = output.repeat(1, 3, 1, 1)
            output = T.ColorJitter(brightness=0.4, contrast=0.3, saturation=0.3, hue=0.2)(output)
            output = T.Grayscale()(output)

    return output.detach().clone(), grid, grid_inverse


def get_gaussian_kernel(kernlen=21, nsig=5):
    """Get kernels used for generating Gaussian heatmaps"""
    interval = (2 * nsig + 1.) / kernlen
    x = np.linspace(-nsig - interval / 2., nsig + interval / 2., kernlen + 1)

    kern1d = np.diff(st.norm.cdf(x))

    kernel_raw = np.sqrt(np.outer(kern1d, kern1d))
    kernel = kernel_raw / kernel_raw.sum()
    kernel = torch.FloatTensor(kernel).unsqueeze(0).unsqueeze(0)
    weight = torch.nn.Parameter(data=kernel, requires_grad=False)
    weight = (weight - weight.min()) / (weight.max() - weight.min())
    return weight


def value_map_load(save_dir, names, input_with_label, shape=(768, 768), value_maps_running=None):
    value_maps = []
    for s, name in enumerate(names):
        path = os.path.join(save_dir, name.split('.')[0] + '.png')
        if input_with_label[s] and value_maps_running is not None and name in value_maps_running:
            value_map = value_maps_running[name]
        elif input_with_label[s] and value_maps_running is None and os.path.exists(path):
            value_map = cv2.imread(path, cv2.IMREAD_GRAYSCALE)
        else:
            value_map = np.zeros([shape[0], shape[1]]).astype(np.uint8)

        value_map = torch.from_numpy(value_map).unsqueeze(0).unsqueeze(0)
        value_maps.append(value_map)
    return torch.cat(value_maps)


def value_map_save(save_dir, names, input_with_label, value_maps, value_maps_running=None):
    for s, name in enumerate(names):
        if input_with_label[s]:
            vp = value_maps[s].squeeze().numpy()
            if value_maps_running is not None:
                value_maps_running[name] = vp
            else:
                path = os.path.join(save_dir, name.split('.')[0] + '.png')
                cv2.imwrite(path, vp)


def train_model(model, optimizer, dataloaders, device, num_epochs, train_config):
    model_save_path = train_config['model_save_path']
    model_save_epoch = train_config['model_save_epoch']
    pke_start_epoch = train_config['pke_start_epoch']
    pke_show_epoch = train_config['pke_show_epoch']
    pke_show_list = train_config['pke_show_list']
    is_value_map_save = train_config['is_value_map_save']
    value_map_save_dir = train_config['value_map_save_dir']

    if is_value_map_save:
        if os.path.exists(value_map_save_dir):
            shutil.rmtree(value_map_save_dir)
        os.makedirs(value_map_save_dir)

    value_maps_running = None
    if not is_value_map_save:
        value_maps_running = {}

    for epoch in range(num_epochs):
        print('Epoch {}/{}'.format(epoch, num_epochs - 1))

        # Each epoch may have some phases
        phases = list(dataloaders.keys())
        model.PKE_learn = True
        if epoch < pke_start_epoch:
            model.PKE_learn = False
        for phase in phases:
            pke_show = train_config['pke_show']
            image_shows = []
            init_kp_shows = []
            label_shows = []
            enhanced_kp_shows = []
            show_names = []

            if 'val' in phase:
                if epoch % model_save_epoch == 0:
                    print(f'save model for epoch {epoch}')
                    state = {'net': model.state_dict(), 'optimizer': optimizer.state_dict(), 'epoch': epoch}
                    torch.save(state, model_save_path)
                continue
            print('-' * 10 + 'phase:' + phase + '\t PKE_learn:' + str(model.PKE_learn) + '-' * 10)

            if 'train' in phase:
                for param_group in optimizer.param_groups:
                    print("LR", param_group['lr'])

                model.train()  # Set model to training mode
            else:
                model.eval()  # Set model to evaluate mode

            epoch_samples = 0
            print_descriptor_loss = 0
            print_detector_loss = 0
            num_learned_pts = 0
            num_input_with_label = 0
            num_input_descriptor = 0

            for images, input_with_label, keypoint_positions, label_names \
                    in tqdm(dataloaders[phase]):
                batch_size = images.shape[0]
                learn_index = torch.where(input_with_label)

                images = images.to(device)

                value_maps = value_map_load(value_map_save_dir, label_names, input_with_label,
                                            images.shape[-2:], value_maps_running)

                value_maps = value_maps.to(device)
                keypoint_positions = keypoint_positions.to(device)

                optimizer.zero_grad()

                # track history if only in train
                with torch.set_grad_enabled('train' in phase):

                    loss, number_pts_one, print_loss_detector_one, print_loss_descriptor_one, enhanced_label_pts, \
                        enhanced_label, detector_pred, loss_detector_num, loss_descriptor_num \
                        = model(images, keypoint_positions, value_maps, learn_index)
                    if enhanced_label_pts is None:
                        enhanced_label_pts = keypoint_positions
                    if pke_show:
                        if len(pke_show_list) == 0 and len(learn_index[0]) != 0:  # randomly show one image
                            show_names = [(label_names[learn_index[0][0]], learn_index[0][0])]
                            pke_show = False
                        for s, label_path in enumerate(label_names):
                            file_name = os.path.split(label_path)[-1].split('.')[0]
                            if file_name in pke_show_list:
                                show_names.append((label_path, s))

                        for (show_name, idx) in show_names:
                            image_shows.append(images[idx][0].cpu().data)
                            init_kp_shows.append(keypoint_positions[idx][0].cpu().data)
                            label_shows.append(enhanced_label[idx][0].cpu().data)
                            pred_show = detector_pred.detach().cpu()[idx][0]
                            enhanced_kp_shows.append(enhanced_label_pts[idx][0].cpu().data)

                    print_detector_loss += print_loss_detector_one * len(learn_index[0])
                    print_descriptor_loss += print_loss_descriptor_one * batch_size
                    num_input_with_label += loss_detector_num

                    num_learned_pts += number_pts_one

                    if 'train' in phase:
                        loss.backward()
                        optimizer.step()
                # update value map
                if len(learn_index[0]) != 0:
                    value_maps = value_maps.cpu()
                    value_map_save(value_map_save_dir, label_names, input_with_label, value_maps, value_maps_running)

                # statistics
                num_input_descriptor += loss_descriptor_num
                epoch_samples += batch_size

                ###################################################################################
            print_detector_loss = print_detector_loss / num_input_with_label
            print_descriptor_loss = print_descriptor_loss / epoch_samples
            print(phase, 'overall loss: {}'.format(print_detector_loss + print_descriptor_loss),
                  'detector_loss: {} of {} nums, #avg learned keypoints:{} '.format(print_detector_loss,
                                                                                    num_input_with_label,
                                                                                    num_learned_pts / num_input_with_label),
                  'descriptor_loss: {} of {} nums'.format(print_descriptor_loss, num_input_descriptor))

            for s, (name, _) in enumerate(show_names):
                if not epoch % pke_show_epoch == 0:
                    break
                input_show = image_shows[s]
                label_show = label_shows[s]
                init_kp_show = init_kp_shows[s]
                enhanced_kp_show = enhanced_kp_shows[s]
                name = os.path.split(name)[-1].split('.')[0]

                plt.figure(dpi=100)
                plt.imshow(input_show, 'gray')
                plt.title('epoch:{}, phase:{}, name:{}'.format(epoch, phase, name))
                plt.axis('off')
                try:
                    y, x = torch.where(enhanced_kp_show.cpu() == 1)
                    plt.scatter(x, y, s=2, c='springgreen')
                except Exception:
                    pass
                try:
                    y, x = torch.where(init_kp_show.cpu() == 1)
                    plt.scatter(x, y, s=2, c='b')
                except Exception:
                    pass
                # plt.savefig(f'./data/draw/train_new_new/{name1}_epo{epoch}_{len(x)}.png',bbox_inches='tight',pad_inches = -0.1)
                plt.show()
                plt.figure(dpi=200)
                plt.subplot(121)
                plt.imshow(pred_show, 'gray')
                plt.subplot(122)
                plt.imshow(label_show, 'gray')
                plt.show()
                plt.close()
                plt.pause(0.1)



import unittest
import os
import numpy as np
import torch
import matplotlib.pyplot as plt
from fundus_image_toolbox.circle_crop import crop

DIR = os.path.join(os.path.dirname(__file__))
fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")


class TestCircleCrop(unittest.TestCase):
    def setUp(self):
        self.image_path = fundus1_path
        self.image = plt.imread(self.image_path)
        self.size = (100, 100)
        self.shape = self.size + (3,)

    def test_circle_crop_with_size(self):
        cropped_image = crop(self.image, size=self.size)
        self.assertEqual(cropped_image.shape, self.shape)

    def test_circle_crop_with_tensor(self):
        tensor_image = torch.from_numpy(self.image.copy())
        cropped_image = crop(tensor_image, to_numpy=False)
        self.assertIsInstance(cropped_image, torch.Tensor)

    def test_circle_crop_returnall(self):
        # Test the circle_crop function
        img_aligned, mask, center, radius = crop(
            self.image, size=self.size, return_all=True, to_numpy=True
        )

        # Assert the output types
        self.assertIsInstance(img_aligned, np.ndarray)
        self.assertIsInstance(mask, np.ndarray)
        self.assertIsInstance(center, np.ndarray)
        self.assertIsInstance(float(center[0]), float)
        self.assertIsInstance(float(center[1]), float)
        self.assertIsInstance(float(radius), float)

        # Assert the output shapes and values
        self.assertEqual(img_aligned.shape, self.shape)
        self.assertEqual(mask.shape, self.size)
        self.assertEqual(len(center), 2)
        self.assertGreaterEqual(radius, 0)


if __name__ == "__main__":
    unittest.main()



import torch
from torchvision import models

def get_efficientnet_or_resnet(type:str, n_outs:int=1):
    """Get a torchvision EfficientNet or ResNet model initialized with ImageNet weights. 
        Choices: 
        resnet18, resnet34, resnet50, resnet101, resnet152,
        efficientnet-b0, efficientnet-b1, efficientnet-b2, efficientnet-b3, 
        efficientnet-b4, efficientnet-b5, efficientnet-b6, efficientnet-b7

    Args:
        type (str): model type
        n_outs (int): number of outputs

    Returns:
        torch.nn.Module: model
    """
    if type == "resnet18":
        model = models.resnet18(weights="IMAGENET1K_V1")
        model.fc = torch.nn.Linear(512, n_outs)
    elif type == "resnet34":
        model = models.resnet34(weights="IMAGENET1K_V1")
        model.fc = torch.nn.Linear(512, n_outs)
    elif type == "resnet50":
        model = models.resnet50(weights="IMAGENET1K_V2")
        model.fc = torch.nn.Linear(2048, n_outs)
    elif type == "resnet101":
        model = models.resnet101(weights="IMAGENET1K_V2")
        model.fc = torch.nn.Linear(2048, n_outs)
    elif type == "resnet152":
        model = models.resnet152(weights="IMAGENET1K_V2")
        model.fc = torch.nn.Linear(2048, n_outs)
    elif type == "efficientnet-b0":
        model = models.efficientnet_b0(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(1280, n_outs)
    elif type == "efficientnet-b1":
        model = models.efficientnet_b1(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(1280, n_outs)
    elif type == "efficientnet-b2":
        model = models.efficientnet_b2(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(1408, n_outs)
    elif type == "efficientnet-b3":
        model = models.efficientnet_b3(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(1536, n_outs)
    elif type == "efficientnet-b4":
        model = models.efficientnet_b4(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(1792, n_outs)
    elif type == "efficientnet-b5":
        model = models.efficientnet_b5(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(2048, n_outs)
    elif type == "efficientnet-b6":
        model = models.efficientnet_b6(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(2304, n_outs)
    elif type == "efficientnet-b7":
        model = models.efficientnet_b7(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(2560, n_outs)        

    else:
        raise ValueError("Model type not supported")
    
    return model


import os


def remove_folder(target_dir):
    for root, dirs, files in os.walk(target_dir, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir(target_dir)


def clone_repo(
    link="https://github.com/berenslab/MIDL24-segmentation_quality_control",
    commit="995a173c1c1735b620c14b2268ce3ca408c10ca1",
    branch="main",
    target_dir=f"./segmentation",
):
    # Clone Segmentation Quality Control repo
    os.makedirs(target_dir, exist_ok=True)
    if not os.listdir(target_dir):
        print(
            f"The vessel segmentation folder is missing. Downloading it from {link}..."
        )
        pwd = os.getcwd()
        os.system(f"git clone {link} {target_dir}")
        os.system(
            f"cd {target_dir} && git checkout {branch} && git reset --hard {commit}"
        )
        try:
            remove_folder(os.path.join(target_dir, ".git"))
        except:
            pass
        os.chdir(pwd)

        print("Adjusting imports...")
        add_dots_to_imports_in_folder(target_dir)
        print("Done.")


# Make imports relative
def add_dots_to_imports(file_path):
    with open(file_path, "r") as file:
        lines = file.readlines()
    with open(file_path, "w") as file:
        for line in lines:
            is_from_import = (
                line.split()[0] == "from" if len(line.split()) > 1 else False
            )
            is_import = line.split()[0] == "import" if len(line.split()) > 1 else False
            is_local_file = False
            is_local_folder = False

            if is_from_import or is_import:
                this_dir = os.path.dirname(file_path)
                par_dir = os.path.dirname(this_dir)
                fname = line.split()[1]
                if fname.startswith("."):
                    file.write(line)
                    continue
                fname = fname.split(".")[0]
                # print(fname)

                # Check if is a local file in same directory
                if os.path.exists(os.path.join(this_dir, fname)):
                    print(os.path.join(this_dir, fname + ".py"))
                    is_local_file = True
                # Check if is a local folder in parent directory
                elif os.path.exists(os.path.join(par_dir, fname)):
                    print(os.path.join(par_dir, fname))
                    is_local_folder = True

            c = ".." if is_local_folder else "."
            if is_local_file or is_local_folder:
                terms = (
                    line.replace("import", ".")
                    .replace("from", "")
                    .replace("\n", "")
                    .split(".")
                )
                terms = [t.replace(" ", "") for t in terms]
                parents = terms[:-1] if len(terms) > 1 else []
                imports = terms[-1].split(",")
                imports = [i.replace(" ", "") for i in imports]

                print("parents: ", parents)
                print("imports: ", imports)

                if len(parents) > 0:
                    file.write(
                        f"from {c}{'.'.join(parents)} import {', '.join(imports)}\n"
                    )
                else:
                    file.write(f"from {c} import {', '.join(imports)}\n")

            else:
                file.write(line)


def add_dots_to_imports_in_folder(folder_path):
    for root, _, files in os.walk(folder_path):
        for file in files:
            if file.endswith(".py"):
                print("########\n", file)
                add_dots_to_imports(os.path.join(root, file))



from types import SimpleNamespace
import numpy as np
import pandas as pd
import os
from PIL import Image
import torch
from torch.utils.data import Dataset, DataLoader
from torchvision.transforms import Compose
from sklearn.model_selection import train_test_split

from .transforms_multi import (
    RandomZoom,
    ResizedCenterCropAndPad,
    RandomSqueeze,
    RandomHorizontalFlip,
    RandomVerticalFlip,
    RandomTranslation,
    RandomRotation,
    ImageAdjustment,
    ToTensor,
)
from .default import DEFAULT_CSV_PATH

SPLIT_SEED = 12345


class ODFoveaDataset(Dataset):
    def __init__(
        self, df: pd.DataFrame, split="train", device="cuda:0", data_root=".."
    ):
        self.df = df
        self.transformation = self._get_transformation(split)
        self.device = device
        self.data_root = data_root

    def __getitem__(self, index):
        image_path = self.df.iloc[index]["image_path"]
        # image = Image.open(os.path.join(self.data_root,image_path))
        image = Image.open(self.data_root / image_path)
        fovea_label = self.df.iloc[index][["fovea_x", "fovea_y"]].values.astype(float)
        od_label = self.df.iloc[index][["od_x", "od_y"]].values.astype(float)
        label = (fovea_label, od_label)
        image, label = self.transformation((image, label))
        fovea_label, od_label = label
        return (
            image.to(self.device),
            fovea_label.to(self.device),
            od_label.to(self.device),
        )

    def __len__(self):
        return len(self.df)

    def _get_transformation(self, split):
        if split == "train":
            return Compose(
                [
                    RandomSqueeze(),
                    RandomHorizontalFlip(),
                    RandomVerticalFlip(),
                    RandomTranslation(),
                    RandomRotation(),
                    ResizedCenterCropAndPad(),
                    RandomZoom(),
                    ImageAdjustment(),
                    ToTensor(),
                ]
            )
        elif split in ["val", "test"]:
            return Compose([ResizedCenterCropAndPad(), ToTensor()])
        else:
            raise ValueError("Split should be either 'train' or 'test'")


class ODFoveaLoader:
    def __init__(self, config: SimpleNamespace):
        np.random.seed(config.seed)
        torch.manual_seed(config.seed)
        torch.cuda.manual_seed(config.seed)

        if config.csv_path is None:
            config.csv_path = DEFAULT_CSV_PATH
        if config.data_root is None:
            config.data_root = "../"

        self.df = self.get_df(config.csv_path)

        # Split into 20 percent validation, 20 percent test
        dev_df, self.test_df = train_test_split(
            self.df, test_size=0.2, shuffle=True, random_state=SPLIT_SEED
        )
        self.train_df, self.val_df = train_test_split(
            dev_df, test_size=0.25, shuffle=True, random_state=SPLIT_SEED
        )

        self.train_df = self.train_df.reset_index(drop=True)
        self.val_df = self.val_df.reset_index(drop=True)
        self.test_df = self.test_df.reset_index(drop=True)

        self.train_dataset = ODFoveaDataset(
            self.train_df,
            split="train",
            device=config.device,
            data_root=config.data_root,
        )
        self.val_dataset = ODFoveaDataset(
            self.val_df, split="val", device=config.device, data_root=config.data_root
        )
        self.test_dataset = ODFoveaDataset(
            self.test_df, split="test", device=config.device, data_root=config.data_root
        )

        print(f"Train dataset: {len(self.train_dataset)}")
        print(f"Val dataset: {len(self.val_dataset)}")
        print(f"Test dataset: {len(self.test_dataset)}")

        self.train_loader = DataLoader(
            self.train_dataset, batch_size=config.batch_size, shuffle=False
        )
        self.val_loader = DataLoader(
            self.val_dataset, batch_size=config.batch_size, shuffle=False
        )
        self.test_loader = DataLoader(
            self.test_dataset, batch_size=config.batch_size, shuffle=False
        )

    def get_dataloaders(self):
        return self.train_loader, self.val_loader, self.test_loader

    def get_datasets(self):
        return self.train_dataset, self.val_dataset, self.test_dataset

    def get_df(self, path):
        df = pd.read_csv(path).reset_index(drop=True)
        return df



import unittest
import os
import numpy as np
import torch
from PIL import Image
from torchvision.transforms import Compose, ToTensor, Resize
from fundus_image_toolbox.quality_prediction import (
    load_quality_ensemble,
    ensemble_predict_quality,
    plot_quality,
    FundusQualityModel,
)

DIR = os.path.join(os.path.dirname(__file__))

fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")


class TestQualityPrediction(unittest.TestCase):
    def setUp(self):
        # Initialize test variables here
        self.image_path = fundus1_path
        self.transforms = Compose([ToTensor(), Resize((350, 350))])
        self.image = self.transforms(Image.open(self.image_path)).unsqueeze(0)
        self.model = load_quality_ensemble(device="cpu")

    def test_load_quality_ensemble(self):
        # Test the load_quality_ensemble function
        for model in self.model:
            self.assertIsInstance(model, FundusQualityModel)

    def test_ensemble_predict_quality(self):
        # Test the ensemble_predict_quality function
        conf, label = ensemble_predict_quality(self.model, self.image)
        self.assertIsInstance(float(conf), float)
        self.assertIsInstance(float(label), float)

    def test_plot_quality(self):
        # Test the plot_quality function
        plot_quality(self.image.squeeze(), 0.5, 1)


if __name__ == "__main__":
    unittest.main()



import configparser
import torch
import cv2
import numpy as np
import matplotlib.pyplot as plt
from torchvision import transforms

from .common.common_util import (
    pre_processing,
    simple_nms,
    remove_borders,
    sample_keypoint_desc,
)
from .model.super_retina import SuperRetina

from PIL import Image
import os


class Predictor:
    def __init__(self, config):

        predict_config = config["PREDICT"]

        device = predict_config["device"]
        device = torch.device(device if torch.cuda.is_available() else "cpu")

        model_save_path = predict_config["model_save_path"]
        self.nms_size = predict_config["nms_size"]
        self.nms_thresh = predict_config["nms_thresh"]
        self.scale = 8
        self.knn_thresh = predict_config["knn_thresh"]

        self.image_width = None
        self.image_height = None

        self.model_image_width = predict_config["model_image_width"]
        self.model_image_height = predict_config["model_image_height"]

        checkpoint = torch.load(model_save_path, map_location=device, weights_only=True)
        model = SuperRetina()
        model.load_state_dict(checkpoint["net"])
        model.to(device)
        model.eval()
        self.device = device
        self.model = model
        self.knn_matcher = cv2.BFMatcher(cv2.NORM_L2)

        self.trasformer = transforms.Compose(
            [
                transforms.Resize((self.model_image_height, self.model_image_width)),
                transforms.ToTensor(),
            ]
        )

    def image_read(self, query_path, refer_path, query_is_image=False):
        if query_is_image:
            query_image = query_path
        else:
            query_image = cv2.imread(query_path, cv2.IMREAD_COLOR)
            # green channel
            query_image = query_image[:, :, 1]
            query_image = pre_processing(query_image)
        refer_image = cv2.imread(refer_path, cv2.IMREAD_COLOR)

        assert query_image.shape[:2] == refer_image.shape[:2]
        self.image_height, self.image_width = query_image.shape[:2]

        refer_image = refer_image[:, :, 1]
        refer_image = pre_processing(refer_image)

        query_image = (query_image * 255).astype(np.uint8)
        refer_image = (refer_image * 255).astype(np.uint8)

        return query_image, refer_image

    def draw_result(
        self, query_image, refer_image, cv_kpts_query, cv_kpts_refer, matches, status
    ):
        def drawMatches(imageA, imageB, kpsA, kpsB, matches, status):
            # initialize the output visualization image
            (hA, wA) = imageA.shape[:2]
            (hB, wB) = imageB.shape[:2]
            vis = np.zeros((max(hA, hB), wA + wB, 3), dtype="uint8")
            if len(imageA.shape) == 2:
                imageA = cv2.cvtColor(imageA, cv2.COLOR_GRAY2RGB)
                imageB = cv2.cvtColor(imageB, cv2.COLOR_GRAY2RGB)

            vis[0:hA, 0:wA] = imageA
            vis[0:hB, wA:] = imageB

            # loop over the matches
            for (match, _), s in zip(matches, status):
                trainIdx, queryIdx = match.trainIdx, match.queryIdx
                # only process the match if the keypoint was successfully
                # matched
                if s == 1:
                    # draw the match
                    ptA = (int(kpsA[queryIdx].pt[0]), int(kpsA[queryIdx].pt[1]))
                    ptB = (int(kpsB[trainIdx].pt[0]) + wA, int(kpsB[trainIdx].pt[1]))
                    cv2.line(vis, ptA, ptB, (0, 255, 0), 2)

                # return the visualization
            return vis

        query_np = np.array([kp.pt for kp in cv_kpts_query])
        refer_np = np.array([kp.pt for kp in cv_kpts_refer])
        refer_np[:, 0] += query_image.shape[1]
        matched_image = drawMatches(
            query_image, refer_image, cv_kpts_query, cv_kpts_refer, matches, status
        )
        plt.figure(dpi=300)
        plt.scatter(query_np[:, 0], query_np[:, 1], s=1, c="r")
        plt.scatter(refer_np[:, 0], refer_np[:, 1], s=1, c="r")
        plt.axis("off")
        plt.title("Match Result, #goodMatch: {}".format(status.sum()))
        plt.imshow(cv2.cvtColor(matched_image, cv2.COLOR_BGR2RGB))
        plt.show()
        plt.close()

    def model_run_pair(self, query_tensor, refer_tensor):
        inputs = torch.cat((query_tensor.unsqueeze(0), refer_tensor.unsqueeze(0)))
        inputs = inputs.to(self.device)

        with torch.no_grad():
            detector_pred, descriptor_pred = self.model(inputs)

        scores = simple_nms(detector_pred, self.nms_size)

        b, _, h, w = detector_pred.shape
        scores = scores.reshape(-1, h, w)

        keypoints = [torch.nonzero(s > self.nms_thresh) for s in scores]

        scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]

        # Discard keypoints near the image borders
        keypoints, scores = list(
            zip(*[remove_borders(k, s, 4, h, w) for k, s in zip(keypoints, scores)])
        )

        keypoints = [torch.flip(k, [1]).float().data for k in keypoints]

        descriptors = [
            sample_keypoint_desc(k[None], d[None], 8)[0].cpu()
            for k, d in zip(keypoints, descriptor_pred)
        ]
        keypoints = [k.cpu() for k in keypoints]
        return keypoints, descriptors

    def match(self, query_path, refer_path, show=False, query_is_image=False):
        query_image, refer_image = self.image_read(
            query_path, refer_path, query_is_image
        )
        query_tensor = self.trasformer(Image.fromarray(query_image))
        refer_tensor = self.trasformer(Image.fromarray(refer_image))

        keypoints, descriptors = self.model_run_pair(query_tensor, refer_tensor)

        query_keypoints, refer_keypoints = keypoints[0], keypoints[1]
        query_desc, refer_desc = (
            descriptors[0].permute(1, 0).numpy(),
            descriptors[1].permute(1, 0).numpy(),
        )

        # mapping keypoints to scaled keypoints
        cv_kpts_query = [
            cv2.KeyPoint(
                int(i[0] / self.model_image_width * self.image_width),
                int(i[1] / self.model_image_height * self.image_height),
                30,
            )
            for i in query_keypoints
        ]
        cv_kpts_refer = [
            cv2.KeyPoint(
                int(i[0] / self.model_image_width * self.image_width),
                int(i[1] / self.model_image_height * self.image_height),
                30,
            )
            for i in refer_keypoints
        ]

        goodMatch = []
        status = []
        matches = []
        try:
            matches = self.knn_matcher.knnMatch(query_desc, refer_desc, k=2)
            for m, n in matches:
                if m.distance < self.knn_thresh * n.distance:
                    goodMatch.append(m)
                    status.append(True)
                else:
                    status.append(False)
        except Exception:
            pass

        if show:
            self.draw_result(
                query_image,
                refer_image,
                cv_kpts_query,
                cv_kpts_refer,
                matches,
                np.array(status),
            )
        return goodMatch, cv_kpts_query, cv_kpts_refer, query_image, refer_image

    def compute_homography(self, query_path, refer_path, query_is_image=False):
        goodMatch, cv_kpts_query, cv_kpts_refer, raw_query_image, raw_refer_image = (
            self.match(query_path, refer_path, query_is_image=query_is_image)
        )
        H_m = None
        inliers_num_rate = 0

        if len(goodMatch) >= 4:
            src_pts = [cv_kpts_query[m.queryIdx].pt for m in goodMatch]
            src_pts = np.float32(src_pts).reshape(-1, 1, 2)
            dst_pts = [cv_kpts_refer[m.trainIdx].pt for m in goodMatch]
            dst_pts = np.float32(dst_pts).reshape(-1, 1, 2)

            H_m, mask = cv2.findHomography(src_pts, dst_pts, cv2.LMEDS)

            # src_pts = src_pts[mask.ravel() == 1]
            # dst_pts = dst_pts[mask.ravel() == 1]

            goodMatch = np.array(goodMatch)[mask.ravel() == 1]
            inliers_num_rate = mask.sum() / len(mask.ravel())

        return H_m, inliers_num_rate, raw_query_image, raw_refer_image

    def align_image_pair(self, query_path, refer_path, show=False):
        H_m, inliers_num_rate, raw_query_image, raw_refer_image = (
            self.compute_homography(query_path, refer_path)
        )

        if H_m is not None:
            h, w = self.image_height, self.image_width
            query_align = cv2.warpPerspective(
                raw_query_image,
                H_m,
                (w, h),
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=(0),
            )

            merged = np.zeros((h, w, 3), dtype=np.uint8)

            if len(query_align.shape) == 3:
                query_align = cv2.cvtColor(query_align, cv2.COLOR_BGR2GRAY)
            if len(raw_refer_image.shape) == 3:
                refer_gray = cv2.cvtColor(raw_refer_image, cv2.COLOR_BGR2GRAY)
            else:
                refer_gray = raw_refer_image
            merged[:, :, 0] = query_align
            merged[:, :, 1] = refer_gray

            if show:
                plt.figure(dpi=200)
                plt.imshow(merged)
                plt.axis("off")
                plt.title("Registration Result")
                plt.show()
                plt.close()
            return merged

        print("Matched Failed!")

    def model_run_one_image(self, image_path, save_path=None):
        image = cv2.imread(image_path, cv2.IMREAD_COLOR)
        image = image[:, :, 1]
        self.image_height, self.image_width = image.shape[:2]

        image = pre_processing(image)
        image_tensor = self.trasformer(Image.fromarray(image))
        inputs = image_tensor.unsqueeze(0)
        inputs = inputs.to(self.device)

        with torch.no_grad():
            detector_pred, descriptor_pred = self.model(inputs)

        scores = simple_nms(detector_pred, self.nms_size)

        b, _, h, w = detector_pred.shape
        scores = scores.reshape(-1, h, w)

        keypoints = [torch.nonzero(s > self.nms_thresh) for s in scores]

        scores = [s[tuple(k.t())] for s, k in zip(scores, keypoints)]

        # Discard keypoints near the image borders
        keypoints, scores = list(
            zip(*[remove_borders(k, s, 4, h, w) for k, s in zip(keypoints, scores)])
        )

        keypoints = [torch.flip(k, [1]).float().data for k in keypoints]

        descriptors = [
            sample_keypoint_desc(k[None], d[None], 8)[0].cpu()
            for k, d in zip(keypoints, descriptor_pred)
        ]
        keypoints = [k.cpu() for k in keypoints]

        if save_path is not None:
            save_info = {"kp": keypoints[0].cpu(), "desc": descriptors[0].cpu()}
            torch.save(save_info, save_path)

        return keypoints[0], descriptors[0]

    def homography_from_tensor(self, query_info, refer_info):
        query_keypoints, query_desc = query_info["kp"], query_info["desc"]
        refer_keypoints, refer_desc = refer_info["kp"], refer_info["desc"]

        query_desc = query_desc.permute(1, 0).numpy()
        refer_desc = refer_desc.permute(1, 0).numpy()
        cv_kpts_query = [
            cv2.KeyPoint(
                int(i[0] / self.model_image_width * self.image_width),
                int(i[1] / self.model_image_height * self.image_height),
                30,
            )
            for i in query_keypoints
        ]
        cv_kpts_refer = [
            cv2.KeyPoint(
                int(i[0] / self.model_image_width * self.image_width),
                int(i[1] / self.model_image_height * self.image_height),
                30,
            )
            for i in refer_keypoints
        ]

        goodMatch = []
        status = []
        try:
            matches = self.knn_matcher.knnMatch(query_desc, refer_desc, k=2)
            for m, n in matches:
                if m.distance < self.knn_thresh * n.distance:
                    goodMatch.append(m)
                    status.append(True)
                else:
                    status.append(False)
        except Exception:
            pass

        H_m = None
        inliers_num = 0

        if len(goodMatch) >= 4:
            src_pts = [cv_kpts_query[m.queryIdx].pt for m in goodMatch]
            src_pts = np.float32(src_pts).reshape(-1, 1, 2)
            dst_pts = [cv_kpts_refer[m.trainIdx].pt for m in goodMatch]
            dst_pts = np.float32(dst_pts).reshape(-1, 1, 2)

            H_m, mask = cv2.findHomography(src_pts, dst_pts, cv2.LMEDS)

            # src_pts = src_pts[mask.ravel() == 1]
            # dst_pts = dst_pts[mask.ravel() == 1]

            goodMatch = np.array(goodMatch)[mask.ravel() == 1]
            inliers_num = mask.sum()

        return H_m, inliers_num


if __name__ == "__main__":
    import yaml

    config_path = "config/test.yaml"
    if os.path.exists(config_path):
        with open(config_path) as f:
            config = yaml.safe_load(f)
    else:
        raise FileNotFoundError("Config File doesn't Exist")

    P = Predictor(config)
    f1 = "./data/samples/query.jpg"
    f2 = "./data/samples/refer.jpg"
    P.match(f1, f2, show=True)
    merged = P.align_image_pair(f1, f2)
    plt.imshow(merged)
    plt.show()



import torch


def pairwise_distance(x1, x2, p=2, eps=1e-6):
    r"""
    Computes the batchwise pairwise distance between vectors v1,v2:
        .. math ::
            \Vert x \Vert _p := \left( \sum_{i=1}^n  \vert x_i \vert ^ p \right) ^ {1/p}
        Args:
            x1: first input tensor
            x2: second input tensor
            p: the norm degree. Default: 2
        Shape:
            - Input: :math:`(N, D)` where `D = vector dimension`
            - Output: :math:`(N, 1)`
        >>> input1 = autograd.Variable(torch.randn(100, 128))
        >>> input2 = autograd.Variable(torch.randn(100, 128))
        >>> output = F.pairwise_distance(input1, input2, p=2)
        >>> output.backward()
    """
    assert x1.size() == x2.size(), "Input sizes must be equal."
    assert x1.dim() == 2, "Input must be a 2D matrix."

    return 1 - torch.cosine_similarity(x1, x2, dim=1)
    # diff = torch.abs(x1 - x2)
    # out = torch.sum(torch.pow(diff + eps, p), dim=1)
    #
    # return torch.pow(out, 1. / p)


def triplet_margin_loss_gor_one(anchor, positive, negative, beta=1.0, margin=1.0, p=2, eps=1e-6, swap=False):
    assert anchor.size() == positive.size(), "Input sizes between positive and negative must be equal."
    assert anchor.size() == negative.size(), "Input sizes between anchor and negative must be equal."
    assert positive.size() == negative.size(), "Input sizes between positive and negative must be equal."
    assert anchor.dim() == 2, "Inputd must be a 2D matrix."
    assert margin > 0.0, 'Margin should be positive value.'
    d_p = pairwise_distance(anchor, positive, p, eps)
    d_n = pairwise_distance(anchor, negative, p, eps)

    dist_hinge = torch.clamp(margin + d_p - d_n, min=0.0)

    neg_dis = torch.pow(torch.sum(torch.mul(anchor, negative), 1), 2)
    gor = torch.mean(neg_dis)

    loss = torch.mean(dist_hinge) + beta * (gor)

    return loss


def triplet_margin_loss_gor(anchor, positive, negative1, negative2, beta=1.0, margin=1.0, p=2, eps=1e-6, swap=False):
    assert anchor.size() == positive.size(), "Input sizes between positive and negative must be equal."
    assert anchor.size() == negative1.size(), "Input sizes between anchor and negative must be equal."
    assert positive.size() == negative2.size(), "Input sizes between positive and negative must be equal."
    assert anchor.dim() == 2, "Inputd must be a 2D matrix."
    assert margin > 0.0, 'Margin should be positive value.'

    # loss1 = triplet_margin_loss_gor_one(anchor, positive, negative1)
    # loss2 = triplet_margin_loss_gor_one(anchor, positive, negative2)
    #
    # return 0.5*(loss1+loss2)

    d_p = pairwise_distance(anchor, positive, p, eps)
    d_n1 = pairwise_distance(anchor, negative1, p, eps)
    d_n2 = pairwise_distance(anchor, negative2, p, eps)

    dist_hinge = torch.clamp(margin + d_p - 0.5 * (d_n1 + d_n2), min=0.0)

    neg_dis1 = torch.pow(torch.sum(torch.mul(anchor, negative1), 1), 2)
    gor1 = torch.mean(neg_dis1)
    neg_dis2 = torch.pow(torch.sum(torch.mul(anchor, negative2), 1), 2)
    gor2 = torch.mean(neg_dis2)

    loss = torch.mean(dist_hinge) + beta * (gor1 + gor2)

    return loss


def distance_matrix_vector(anchor, positive):
    """Given batch of anchor descriptors and positive descriptors calculate distance matrix"""
    D = anchor.shape[-1]
    d1_sq = torch.sum(anchor * anchor, dim=1).unsqueeze(-1)
    d2_sq = torch.sum(positive * positive, dim=1).unsqueeze(-1)

    eps = 1e-3
    return torch.sqrt((d1_sq.repeat(1, positive.size(0)) + torch.t(d2_sq.repeat(1, anchor.size(0)))
                       - 2.0 * torch.bmm(anchor.unsqueeze(0), torch.t(positive).unsqueeze(0)).squeeze(0))+eps)

    # anchor = anchor.permute(1, 0).view(D, -1, 1)
    # positive = positive.permute(1, 0).view(D, 1, -1)
    # return torch.norm(anchor - positive, dim=0)


def percentile(t, q):
    """
    Return the ``q``-th percentile of the flattened input tensor's data.

    CAUTION:
     * Needs PyTorch >= 1.1.0, as ``torch.kthvalue()`` is used.
     * Values are not interpolated, which corresponds to
       ``numpy.percentile(..., interpolation="nearest")``.

    :param t: Input tensor.
    :param q: Percentile to compute, which must be between 0 and 100 inclusive.
    :return: Resulting value (scalar).
    """
    # Note that ``kthvalue()`` works one-based, i.e. the first sorted value
    # indeed corresponds to k=1, not k=0! Use float(q) instead of q directly,
    # so that ``round()`` returns an integer, even if q is a np.float32.
    k = 1 + round(.01 * float(q) * (t.numel() - 1))
    result = t.view(-1).kthvalue(int(k)).values.item()
    return result


""" Triplet loss usd in SOSNet """
def sos_reg(anchor, positive, KNN=True, k=1, eps=1e-8):
    dist_matrix_a = distance_matrix_vector(anchor, anchor) + eps
    dist_matrix_b = distance_matrix_vector(positive, positive) + eps
    if KNN:
        k_max = percentile(dist_matrix_b, k)
        #print("k_max:", k_max)
        mask = dist_matrix_b.lt(k_max)
        dist_matrix_a = dist_matrix_a*mask.int().float()
        dist_matrix_b = dist_matrix_b*mask.int().float()
    SOS_temp = torch.sqrt(torch.sum(torch.pow(dist_matrix_a-dist_matrix_b, 2)))
    return torch.mean(SOS_temp)



import numpy as np


# Compute AUC scores for image registration on the FIRE dataset
def compute_auc(s_error, p_error, a_error):
    assert (len(s_error) == 71)  # Easy pairs
    assert (len(p_error) == 48)  # Hard pairs. Note file control_points_P37_1_2.txt is ignored
    assert (len(a_error) == 14)  # Moderate pairs

    s_error = np.array(s_error)
    p_error = np.array(p_error)
    a_error = np.array(a_error)

    limit = 25
    gs_error = np.zeros(limit + 1)
    gp_error = np.zeros(limit + 1)
    ga_error = np.zeros(limit + 1)

    accum_s = 0
    accum_p = 0
    accum_a = 0

    for i in range(1, limit + 1):
        gs_error[i] = np.sum(s_error < i) * 100 / len(s_error)
        gp_error[i] = np.sum(p_error < i) * 100 / len(p_error)
        ga_error[i] = np.sum(a_error < i) * 100 / len(a_error)

        accum_s = accum_s + gs_error[i]
        accum_p = accum_p + gp_error[i]
        accum_a = accum_a + ga_error[i]

    auc_s = accum_s / (limit * 100)
    auc_p = accum_p / (limit * 100)
    auc_a = accum_a / (limit * 100)
    mAUC = (auc_s + auc_p + auc_a) / 3.0
    return {'s': auc_s, 'p': auc_p, 'a': auc_a, 'mAUC': mAUC}



from .fundus_transforms import get_transforms, get_normalization, get_unnormalization
from .balancing import ImbalancedDatasetSampler
from .set_seed import set_seed as seed_everything
from .get_pixel_mean_std import get_pixel_mean_and_sd
from .get_efficientnet_resnet import get_efficientnet_or_resnet
from .multilevel_3way_split import multilevel_3way_split
from .basics import exists, flatten_one, parse_list, on_slurm_job, show, print_type
from .lr_scheduler import get_lr_scheduler
from .image_torch_utils import ImageTorchUtils


import os
import unittest
import numpy as np
import torch
from PIL import Image
from fundus_image_toolbox.vessel_segmentation import (
    load_segmentation_ensemble,
    ensemble_predict_segmentation,
    plot_masks,
    save_masks,
    load_masks_from_filenames,
    FR_UNet,
)

DIR = os.path.join(os.path.dirname(__file__))

fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")


class TestVesselSegmentation(unittest.TestCase):
    def setUp(self):
        # Initialize test variables here
        self.image = Image.open(fundus1_path)
        self.device = "cpu"
        self.model = load_segmentation_ensemble(device=self.device)

    def test_load_segmentation_ensemble(self):
        # Test the load_segmentation_ensemble function
        for model in self.model:
            self.assertIsInstance(model, FR_UNet)

    def test_ensemble_predict_segmentation(self):
        # Test the ensemble_predict_segmentation function
        segmentation = ensemble_predict_segmentation(
            self.model, self.image, device=self.device
        )
        self.assertIsInstance(segmentation, np.ndarray)
        self.assertEqual(segmentation.shape, (self.image.height, self.image.width))

    def test_plot_masks(self):
        # Test the plot_masks function
        segmentation = ensemble_predict_segmentation(
            self.model, self.image, device=self.device
        )
        plot_masks(self.image, segmentation)

    def test_save_and_load_masks(self):
        # Test the save_masks and load_masks_from_filenames functions
        target_dir = os.path.join(DIR, "masks")
        mask = ensemble_predict_segmentation(self.model, self.image, device=self.device)
        save_masks(fundus1_path, mask, target_dir)
        loaded_mask = load_masks_from_filenames(fundus1_path, target_dir)
        self.assertIsInstance(loaded_mask, np.ndarray)
        self.assertEqual(loaded_mask.shape, mask.shape)

        # Clean up
        for file in os.listdir(os.path.join(target_dir)):
            os.remove(os.path.join(target_dir, file))


if __name__ == "__main__":
    unittest.main()



from .circle_crop import *
from .fovea_od_localization import *
from .quality_prediction import *
from .registration import *
from .registration import *
from .utilities import *
from .vessel_segmentation import *



# Adopted from https://github.com/ufoym/imbalanced-dataset-sampler/
"""
MIT License

Copyright (c) 2018 Ming

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from typing import Callable

import pandas as pd
import torch
import torch.utils.data
import torchvision


class ImbalancedDatasetSampler(torch.utils.data.sampler.Sampler):
    """Samples elements randomly from a given list of indices for imbalanced dataset.
    An imbalanced dataset sampler for oversampling low frequent classes and undersampling high
    frequent ones. Does not alter length of dataset.


    Args:
        dataset: torch dataset
        method: "balanced". Not yet implemented: "oversampling" or "undersampling"
        labels: a list of labels to balance by
        indices: a list of indices
        num_samples: number of samples to draw
        callback_get_label: a callback-like function which takes two arguments - dataset and index

    Returns:
        Iterable: indices to sample from

    Example:
        ```
        if config.balance_datasets:
            # Balance by original dataset origin (DRIMDB, DeepDRID-ISBI2020)
            sampler = ImbalancedDatasetSampler(self.train_dataset, method="balanced", labels = self.train_dataset.data["dataset"])
            shuffle = False
        else:
            sampler = None
            shuffle = True
        self.train_dataloader = torch.utils.data.DataLoader(self.train_dataset, batch_size=config.batch_size, shuffle=shuffle, sampler=sampler)
        ```
    """

    def __init__(
        self,
        dataset,
        method: str = "balanced",
        labels: list = None,
        indices: list = None,
        num_samples: int = None,
        callback_get_label: Callable = None,
    ):
        # if indices is not provided, all elements in the dataset will be considered
        self.indices = list(range(len(dataset))) if indices is None else indices

        # define custom callback
        self.callback_get_label = callback_get_label

        self.method = method

        # distribution of classes in the dataset
        df = pd.DataFrame()
        df["label"] = self._get_labels(dataset) if labels is None else labels
        df.index = self.indices
        df["indices"] = self.indices
        self.df = df.sort_index()

        self.label_count = df["label"].value_counts()
        label_count_sorted = self.label_count.sort_values()

        self.minority_class = label_count_sorted.keys()[0]
        self.majority_class = label_count_sorted.keys()[-1]

        self.df_minority = df[df["label"] == self.minority_class]
        self.df_majority = df[df["label"] == self.majority_class]

        if method == "balanced":
            self.num_samples = len(self.indices) if num_samples is None else num_samples
        else:
            raise NotImplementedError

    def _get_labels(self, dataset):
        if self.callback_get_label:
            return self.callback_get_label(dataset)
        elif isinstance(dataset, torch.utils.data.TensorDataset):
            return dataset.tensors[1]
        elif isinstance(dataset, torchvision.datasets.MNIST):
            return dataset.train_labels.tolist()
        elif isinstance(dataset, torchvision.datasets.ImageFolder):
            return [x[1] for x in dataset.imgs]
        elif isinstance(dataset, torchvision.datasets.DatasetFolder):
            return dataset.samples[:][1]
        elif isinstance(dataset, torch.utils.data.Subset):
            return dataset.dataset.imgs[:][1]
        elif isinstance(dataset, torch.utils.data.Dataset):
            return dataset.get_labels()
        else:
            raise NotImplementedError

    def __iter__(self):
        if self.method == "balanced":
            weights = 1.0 / self.label_count[self.df["label"]]
            weights = torch.DoubleTensor(weights.to_list())
            return (
                self.indices[i]
                for i in torch.multinomial(weights, self.num_samples, replacement=True)
            )

        else:
            raise NotImplementedError

    def __len__(self):
        return self.num_samples



import os
from setuptools import setup, find_packages

project = os.path.basename(os.path.dirname(os.path.realpath(__file__)))
# about = __import__(project).about

with open("Readme.md", "r", encoding='utf8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name=project,
    version="0.0.1",
    author="Julius Gervelmeyer et al.",
    author_email="Julius.Gervelmeyer@uni-tuebingen.de",
    description=long_description.split('\n')[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url=None,
    packages=find_packages(),
    install_requires=required,
    package_dir={'': '.'},
    python_requires='>=3.9.0, <3.10',
)


import torch

class WarmupLRScheduler():
    def __init__(self, optimizer, warmup_epochs, initial_lr):
        self.epoch = 0
        self.optimizer = optimizer
        self.warmup_epochs = warmup_epochs
        self.initial_lr = initial_lr

    def step(self):
        if self.epoch <= self.warmup_epochs:
            self.epoch += 1
            curr_lr = (self.epoch / self.warmup_epochs) * self.initial_lr
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = curr_lr

    def finished(self):
        return self.epoch >= self.warmup_epochs

def get_lr_scheduler(type:str, optimizer:torch.optim.Optimizer, len_train_loader:int, lr:float, epochs:int, cosine_len=None, warmup_epochs=None, warmup_lr=None):
    """Create a learning rate scheduler.
        Options: - "constant" or None for no schedule, 
                 - "onecycle" for OneCycleLR, 
                 - "cosine", for CosineAnnealingLR,
                 - "cosine_restart" for CosineAnnealingWarmRestarts.

        If using warmup, note: Discard warmup scheduler when loading from checkpoint.
        See below for sample usage.
    
    Args:
        type (str): Type of scheduler.
        optimizer (torch.optim.Optimizer): Optimizer.
        len_train_loader (int): Length of training data loader (number of batches per epoch)
        lr (float): Learning rate.
        epochs (int): total number of epochs.
        cosine_len (int, optional): Length of cosine annealing period. If set to None, will use
            1/2*epochs. Defaults to None.
        warmup_epochs (int, optional): Number of warmup epochs. Defaults to None.
        warmup_lr (float, optional): Warmup learning rate. If set to None, will use 1/10*lr.
            Defaults to None.

    Returns:
        torch.optim.lr_scheduler: Learning rate scheduler.
        torch.optim.lr_scheduler: Warmup learning rate scheduler.
            In case of no warmup scheduler, it is set to None.
    
    Example:
        Store the <type> in self.cfg.scheduler.
        In the training loop, at the beginning of each epoch, run:
        ```
        if self.warmup_scheduler and not self.warmup_scheduler.finished():
                self.warmup_scheduler.step()
        ```
        And after each batch, run:
        ```
        if self.scheduler and "onecycle" in self.cfg.scheduler:
            if not self.warmup_scheduler or self.warmup_scheduler.finished() and self.warmup_scheduler.epoch != self.current_epoch+1:
                self.scheduler.step()
        ```
        And at the end of each epoch, run:
        ```
        if self.scheduler and not "onecycle" in self.cfg.scheduler:
            if not self.warmup_scheduler or self.warmup_scheduler.finished():
                self.scheduler.step()
        ```
    """
    type = str(type).lower()
    if "none" in type or "constant" in type:
            return None, None
    
    elif "onecycle" in type:
        scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer,
            max_lr=lr,
            epochs=epochs,
            steps_per_epoch=len_train_loader,
        )

    elif "cosine" in type and not "restart" in type:
        if cosine_len is not None:
            period = cosine_len
        else:
            period = epochs/2
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=int(period)
        )

    elif "cosine" in type and "restart" in type:
        if cosine_len is not None:
            length = cosine_len
        else:
            length = epochs/2
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer, T_0=length, T_mult=1
        )
        
    else:
        raise ValueError(f"Invalid scheduler: {type}!")

    if warmup_epochs is not None and warmup_epochs > 0:
        if warmup_lr is None:
            warmup_lr = float(lr)/10
        warmup_scheduler = WarmupLRScheduler(
            optimizer, warmup_epochs, warmup_lr
        )
        return scheduler, warmup_scheduler
    
    else:
        return scheduler, None


import os
import unittest
import numpy as np
from PIL import Image
from fundus_image_toolbox.registration import (
    register,
    load_registration_model,
    DEFAULT_CONFIG,
    SuperRetina,
)
from cv2 import BFMatcher

DIR = os.path.join(os.path.dirname(__file__))

fundus1_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus1.jpg")
fundus2_path = os.path.join(DIR, "..", "0_example_usage", "imgs", "fundus2.jpg")


class TestRegistration(unittest.TestCase):
    def setUp(self):
        # Initialize test variables here
        self.image1 = Image.open(fundus1_path)
        self.image2 = Image.open(fundus2_path)
        self.config = DEFAULT_CONFIG
        self.config["device"] = "cpu"
        self.model, self.matcher = load_registration_model(self.config)

    def test_load_registration_model(self):
        # Test the load_registration_model function
        self.assertIsInstance(self.model, SuperRetina)
        self.assertIsInstance(self.matcher, BFMatcher)

    def test_register(self):
        # Test the register function
        registered_image = register(
            self.image1,
            self.image2,
            show=False,
            show_mapping=False,
            config=self.config,
            model=self.model,
            matcher=self.matcher,
        )
        self.assertIsInstance(registered_image, np.ndarray)


if __name__ == "__main__":
    unittest.main()



# Warpper for multi-level-split from https://github.com/lmkoch/multi-level-split/

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

SPLIT_SEED = 12345

def multilevel_3way_split(df:pd.DataFrame, proportions:list=[0.6, 0.2, 0.2], seed:int=SPLIT_SEED, stratify_by:str=None, split_by:str=None, verify_proportions_with_atol:float=None):
    """Split a dataframe into train, validation, and test sets. Options to split by group (i.e. keep groups together) and stratify by label.

    Args:
        df (pd.DataFrame): dataframe to split
        proportions (list, optional): Proportions for train, validation, and test sets. Defaults to [0.6, 0.2, 0.2].
        seed (int, optional): random seed. Defaults to 12345.
        stratify_by (str, optional): column to stratify by. Defaults to None.
        split_by (str, optional): column to group together and split by. Defaults to None.
        verify_proportions_with_atol (float, optional): If not None, verify that the returned proportions are within this tolerance to the `proportions` list. Defaults to None.

    Returns:
        pd.DataFrame: train set
        pd.DataFrame: validation set
        pd.DataFrame: test set
    """
    
    # Split proportions [train, val, test]
    p = proportions

    dev_df, test_df = multilevel_train_test_split(df, df.index, test_split=p[1], seed=seed, stratify_by=stratify_by, split_by=split_by)

    train_df, val_df = multilevel_train_test_split(dev_df, dev_df.index, test_split=p[2]/(p[0]+p[1]), seed=seed, stratify_by=stratify_by, split_by=split_by)

    if verify_proportions_with_atol is not None:
        assert np.isclose(len(train_df)/len(df), p[0], atol=verify_proportions_with_atol), f"fraction is not {p[0]} but is {len(train_df)/len(df)}"
        assert np.isclose(len(val_df)/len(df), p[1], atol=verify_proportions_with_atol), f"fraction is not {p[1]} but is {len(val_df)/len(df)}"
        assert np.isclose(len(test_df)/len(df), p[2], atol=verify_proportions_with_atol), f"fraction is not {p[2]} but is {len(test_df)/len(df)}"

    return train_df, val_df, test_df

# From https://github.com/lmkoch/multi-level-split/
def multilevel_train_test_split(df:pd.DataFrame, 
                                index,
                                split_by=None,
                                stratify_by=None,
                                test_split=0.1,
                                seed=None):
    """Split pandas dataframe into train and test splits. Options to split by
    group (i.e. keep groups together) and stratify by label.

    Args:
        df (pandas dataframe): dataframe to split. Must contain columns index as well as
                     split_by and stratify_by (if not None)
        index (string): name of column that acts as index to dataset
        split_by (str, optional): column to group together and split by. Defaults to None.
        stratify_by (str, optional): column to stratify by. Defaults to None.
        test_split (float, optional): test proportion. Defaults to 0.1.
        seed (int, optional): random seed can be fixed for reproducible splits. 
                              Defaults to None.
    """
    if (index == df.index).all():
        df["_index_"] = index
        index = "_index_"

    if index not in df:
        raise ValueError(f'{index} not in df')
        
    if split_by is None:
        split_by = index
    elif split_by not in df:
        raise ValueError(f'{split_by} not a column in df')

    df_unique = df.drop_duplicates(subset=[split_by])

    if stratify_by is None:
        stratify = None
    else:
        if stratify_by in df:
            stratify = df_unique[stratify_by]
        else:
            raise ValueError(f'{stratify_by} not a column in df')
 
    train_ids, test_ids = train_test_split(df_unique[split_by], 
                                           test_size=test_split,
                                           stratify=stratify,
                                           random_state=seed)
    
    df_train = df.set_index(split_by).drop(test_ids).reset_index()
    df_test = df.set_index(split_by).drop(train_ids).reset_index()

    if index == "_index_":
        del df["_index_"]

    return df_train, df_test


import os
from setuptools import setup, find_packages

root = os.path.dirname(os.path.realpath(__file__))
project = os.path.basename(root)
# about = __import__(project).about

with open("Readme.md", "r", encoding='utf8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name=project,
    version="0.0.1",
    author="Julius Gervelmeyer et al.",
    author_email="Julius.Gervelmeyer@uni-tuebingen.de",
    description=long_description.split('\n')[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url=None,
    packages=find_packages(),
    install_requires=required,
    package_dir={'': '.'},
    python_requires='>=3.9.0, <3.10',
)

## The current setup includes the SuperRetina repo already (May 2024), as the authors allowed
##  any usage of their code. If this changed or you want the latest commit, uncomment this code to
##  - clone the latest commit of the SuperRetina repo
##  - adjust all imports for package usage
##  - download weights from their google drive (as opposed to re-upload on zenodo)
##  and re-install the package with `pip install -e .` in the root directory of this registration project.
#
# # Clone SuperRetina repo
# link = "https://github.com/ruc-aimc-lab/SuperRetina"
# commit = "338f041cc2ce86f39623e7da950b14f33bbc25df"
# branch = "main"
# target_dir = f"./{project}/SuperRetina/"
# if not os.path.exists(target_dir) or not os.listdir(target_dir):
#     pwd = os.getcwd()
#     os.system(f"git clone {link} {target_dir}")
#     os.system(f"cd {target_dir} && git checkout {branch} && git reset --hard {commit}")
#     os.system(f"rm -rf {target_dir}/.git")
#     os.chdir(pwd)
#     print("Repo cloned successfully.")

# # Make imports relative
# def add_dots_to_imports(file_path):
#     with open(file_path, 'r') as file:
#         lines = file.readlines()
#     with open(file_path, 'w') as file:
#         for line in lines:
#             is_from_import = line.split()[0] == 'from' if len(line.split()) > 1 else False
#             is_import = line.split()[0] == 'import' if len(line.split()) > 1 else False
#             is_local_file = False
#             is_local_folder = False
            
#             if is_from_import or is_import:
#                 this_dir = os.path.dirname(file_path)
#                 par_dir = os.path.dirname(this_dir)
#                 fname = line.split()[1]
#                 if fname.startswith('.'):
#                     file.write(line)
#                     continue
#                 fname = fname.split('.')[0]
#                 # print(fname)

#                 # Check if is a local file in same directory
#                 if os.path.exists(os.path.join(this_dir, fname)):
#                     print(os.path.join(this_dir, fname + '.py'))
#                     is_local_file = True
#                 # Check if is a local folder in parent directory
#                 elif os.path.exists(os.path.join(par_dir, fname)):
#                     print(os.path.join(par_dir, fname))
#                     is_local_folder = True

#             c = ".." if is_local_folder else "."
#             if is_local_file or is_local_folder:
#                 terms = line.replace("import", ".").replace("from", "").replace("\n", "").split(".")
#                 terms = [t.replace(" ", "") for t in terms]
#                 parents = terms[:-1] if len(terms) > 1 else []
#                 imports = terms[-1].split(",")
#                 imports = [i.replace(" ", "") for i in imports]

#                 print("parents: ", parents)
#                 print("imports: ", imports)
                
#                 if len(parents) > 0:
#                     file.write(f"from {c}{'.'.join(parents)} import {', '.join(imports)}\n")
#                 else:
#                     file.write(f"from {c} import {', '.join(imports)}\n")
                
#             else:
#                 file.write(line)

# def add_dots_to_imports_in_folder(folder_path):
#     for root, _, files in os.walk(folder_path):
#         for file in files:
#             if file.endswith('.py'):
#                 print("########\n", file)
#                 add_dots_to_imports(os.path.join(root, file))

# print("Adjusting imports...")
# add_dots_to_imports_in_folder(f'./{project}/SuperRetina')
# print("Done.")

# # Download weights
# #   original drive link from repo: https://drive.google.com/drive/folders/1h-MH3wEiN7BoLyMRjF1OAwABKqq6gVFL
# #   retrieved cli link using $$("[data-id]").map((el) => 'https://drive.google.com/uc?id=' + el.getAttribute('data-id')).join(" ")
# target_dir = f"./{project}/SuperRetina/save"
# if not os.path.exists(os.path.join(target_dir,"SuperRetina.pth")):
#     weights = "https://drive.google.com/uc?id=1OL7LhLLRSBW72AHOyAwqjCeKV-XpP2GW"
#     e = os.system(f"gdown {weights} -O {target_dir}")
#     # Use if fetching a folder:
#     # os.system(f"mv {target_dir}/SuperRetina/SuperRetina.pth {target_dir}/SuperRetina.pth")
#     # os.system(f"rm -r {target_dir}/SuperRetina")
#     if e == 0:
#         print("Weights downloaded successfully.")
#     else:
#         print("Weights could not be downloaded. This happens if Google blocked automatic access. Please download the weights manually from the link below and place it in the folder 'registration/SuperRetina/save' with the name 'SuperRetina.pth'.")
#         print("Link: https://drive.google.com/uc?id=1OL7LhLLRSBW72AHOyAwqjCeKV-XpP2GW")


# import os
from pathlib import Path
from types import SimpleNamespace

DEFAULT_MODEL = "2024-05-07 11_13.05"
# DEFAULT_CSV_PATH = os.path.join(os.path.dirname(__file__), "data/ADAM+IDRID+REFUGE_df.csv").__str__()
# MODELS_DIR = os.path.join(os.path.dirname(__file__), "models").__str__()
DEFAULT_CSV_PATH = Path(__file__).parent / "data/ADAM+IDRID+REFUGE_df.csv"
MODELS_DIR = Path(__file__).parent / "models"
DEFAULT_CONFIG = SimpleNamespace(
    **{
        "batch_size": 8,
        "csv_path": DEFAULT_CSV_PATH,
        # "data_root": "../../",
        "data_root": Path(__file__).parent.parent,
        "device": "cuda:0",
        "epochs": 500,
        "img_size": 350,
        "lr": 0.0001,
        "model_type": "efficientnet-b3",
        "seed": 123,
        "testset_eval": True,
    }
)



import os
from types import SimpleNamespace
import numpy as np
import pandas as pd
import torch

from .dataset import FundusQualityDataset
from fundus_image_toolbox.utilities import ImbalancedDatasetSampler
from .default import DATA_ROOT


class FundusQualityLoader:
    def __init__(
        self,
        config,
        root_dir=DATA_ROOT,
        augment=True,
        normalize=True,
        verbose=True,
        drimdb_dir=None,
        deepdrid_dir=None,
    ):
        if isinstance(config, dict):
            config = SimpleNamespace(**config)

        self.root_dir = root_dir

        self.train_dataset = FundusQualityDataset(
            config,
            root_dir,
            split="train",
            augment=augment,
            normalize=normalize,
            verbose=verbose,
            drimdb_dir=drimdb_dir,
            deepdrid_dir=deepdrid_dir,
        )
        self.val_dataset = FundusQualityDataset(
            config,
            root_dir,
            split="val",
            augment=augment,
            normalize=normalize,
            verbose=verbose,
            drimdb_dir=drimdb_dir,
            deepdrid_dir=deepdrid_dir,
        )
        self.test_dataset = FundusQualityDataset(
            config,
            root_dir,
            split="test",
            augment=augment,
            normalize=normalize,
            verbose=verbose,
            drimdb_dir=drimdb_dir,
            deepdrid_dir=deepdrid_dir,
        )

        if config.balance_datasets:
            # Balance by original dataset origin: areds, registration_outs, drimdb, deepdrid
            shuffle = False
            sampler = ImbalancedDatasetSampler(
                self.train_dataset,
                method="balanced",
                labels=self.train_dataset.data["dataset"],
            )
        else:
            shuffle = False
            sampler = None

        self.train_dataloader = torch.utils.data.DataLoader(
            self.train_dataset,
            batch_size=config.batch_size,
            shuffle=shuffle,
            sampler=sampler,
        )
        self.val_dataloader = torch.utils.data.DataLoader(
            self.val_dataset, batch_size=config.batch_size, shuffle=False
        )
        self.test_dataloader = torch.utils.data.DataLoader(
            self.test_dataset, batch_size=config.batch_size, shuffle=False
        )

    def get_dataloaders(self):
        return self.train_dataloader, self.val_dataloader, self.test_dataloader

    def get_datasets(self):
        return self.train_dataset, self.val_dataset, self.test_dataset
